""" Simple test for compatibility with HuggingFace's help text convention.

This checks that Simple-Parsing can be used as a replacement for the HFArgumentParser.
"""
import io
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

import pytest

from simple_parsing import ArgumentParser
from simple_parsing.docstring import get_attribute_docstring

from .testutils import TestSetup, raises_invalid_choice


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune, or train from scratch.
    """

    model_name_or_path: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "The model checkpoint for weights initialization.Don't set if you want to train a model from scratch."
            )
        },
    )
    model_type: Optional[str] = field(
        default=None,
        metadata={
            "help": "If training from scratch, pass a model type from the list: "
            # + ", ".join(MODEL_TYPES)
        },
    )
    config_overrides: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "Override some existing default config settings when a model is trained from scratch. Example: "
                "n_embd=10,resid_pdrop=0.2,scale_attn_weights=false,summary_type=cls_index"
            )
        },
    )
    config_name: Optional[str] = field(
        default=None,
        metadata={"help": "Pretrained config name or path if not the same as model_name"},
    )
    tokenizer_name: Optional[str] = field(
        default=None,
        metadata={"help": "Pretrained tokenizer name or path if not the same as model_name"},
    )
    cache_dir: Optional[str] = field(
        default=None,
        metadata={
            "help": "Where do you want to store the pretrained models downloaded from huggingface.co"
        },
    )
    use_fast_tokenizer: bool = field(
        default=True,
        metadata={
            "help": "Whether to use one of the fast tokenizer (backed by the tokenizers library) or not."
        },
    )
    model_revision: str = field(
        default="main",
        metadata={
            "help": "The specific model version to use (can be a branch name, tag name or commit id)."
        },
    )
    use_auth_token: bool = field(
        default=False,
        metadata={
            "help": (
                "Will use the token generated when running `huggingface-cli login` (necessary to use this script "
                "with private models)."
            )
        },
    )

    def __post_init__(self):
        if self.config_overrides is not None and (
            self.config_name is not None or self.model_name_or_path is not None
        ):
            raise ValueError(
                "--config_overrides can't be used in combination with --config_name or --model_name_or_path"
            )


@dataclass
class DataTrainingArguments:
    """
    Arguments pertaining to what data we are going to input our model for training and eval.
    """

    dataset_name: Optional[str] = field(
        default=None,
        metadata={"help": "The name of the dataset to use (via the datasets library)."},
    )
    dataset_config_name: Optional[str] = field(
        default=None,
        metadata={
            "help": "The configuration name of the dataset to use (via the datasets library)."
        },
    )
    train_file: Optional[str] = field(
        default=None, metadata={"help": "The input training data file (a text file)."}
    )
    validation_file: Optional[str] = field(
        default=None,
        metadata={
            "help": "An optional input evaluation data file to evaluate the perplexity on (a text file)."
        },
    )
    max_train_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "For debugging purposes or quicker training, truncate the number of training examples to this "
                "value if set."
            )
        },
    )
    max_eval_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "For debugging purposes or quicker training, truncate the number of evaluation examples to this "
                "value if set."
            )
        },
    )

    block_size: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "Optional input sequence length after tokenization. "
                "The training dataset will be truncated in block of this size for training. "
                "Default to the model max input length for single sentence inputs (take into account special tokens)."
            )
        },
    )
    overwrite_cache: bool = field(
        default=False,
        metadata={"help": "Overwrite the cached training and evaluation sets"},
    )
    validation_split_percentage: Optional[int] = field(
        default=5,
        metadata={
            "help": "The percentage of the train set used as validation set in case there's no validation split"
        },
    )
    preprocessing_num_workers: Optional[int] = field(
        default=None,
        metadata={"help": "The number of processes to use for the preprocessing."},
    )
    keep_linebreaks: bool = field(
        default=True,
        metadata={"help": "Whether to keep line breaks when using TXT files or not."},
    )

    def __post_init__(self):
        if self.dataset_name is None and self.train_file is None and self.validation_file is None:
            raise ValueError("Need either a dataset name or a training/validation file.")
        else:
            if self.train_file is not None:
                extension = self.train_file.split(".")[-1]
                assert extension in [
                    "csv",
                    "json",
                    "txt",
                ], "`train_file` should be a csv, a json or a txt file."
            if self.validation_file is not None:
                extension = self.validation_file.split(".")[-1]
                assert extension in [
                    "csv",
                    "json",
                    "txt",
                ], "`validation_file` should be a csv, a json or a txt file."


def test_help_text_gets_used():
    parser = ArgumentParser()
    parser.add_arguments(ModelArguments, dest="model_args")
    parser.add_arguments(DataTrainingArguments, dest="data_args")
    with io.StringIO() as f:
        parser.print_help(f)
        f.seek(0)
        help_text = f.read()
    assert "The number of processes to use for the preprocessing." in help_text


@dataclass
class Config(TestSetup):
    log_level: Optional[str] = field(
        default="passive",
        metadata={
            "help": (
                "Logger log level to use on the main node. Possible choices are the log levels as strings: 'debug',"
                " 'info', 'warning', 'error' and 'critical', plus a 'passive' level which doesn't set anything and"
                " lets the application set the level. Defaults to 'passive'."
            ),
            "choices": [
                "debug",
                "info",
                "warning",
                "error",
                "critical",
            ],
        },
    )


def test_choices():
    """Checks that the `choices` in the field metadata are used as the `choice` argument to `add_argument`"""

    with raises_invalid_choice():
        Config.setup("--log_level invalid")
    for choice in ["debug", "info", "warning", "error", "critical"]:
        assert Config.setup(f"--log_level {choice}").log_level == choice


class ExplicitEnum(str, Enum):
    """
    Enum with more explicit error message for missing values.
    """

    @classmethod
    def _missing_(cls, value):
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, please select one of {list(cls._value2member_map_.keys())}"
        )


class IntervalStrategy(ExplicitEnum):
    NO = "no"
    STEPS = "steps"
    EPOCH = "epoch"


class EvaluationStrategy(ExplicitEnum):
    NO = "no"
    STEPS = "steps"
    EPOCH = "epoch"


class HubStrategy(ExplicitEnum):
    END = "end"
    EVERY_SAVE = "every_save"
    CHECKPOINT = "checkpoint"
    ALL_CHECKPOINTS = "all_checkpoints"


class SchedulerType(ExplicitEnum):
    LINEAR = "linear"
    COSINE = "cosine"
    COSINE_WITH_RESTARTS = "cosine_with_restarts"
    POLYNOMIAL = "polynomial"
    CONSTANT = "constant"
    CONSTANT_WITH_WARMUP = "constant_with_warmup"


trainer_log_levels = {
    "debug": -1,
    "info": -1,
    "warning": -1,
    "error": -1,
    "critical": -1,
    "passive": -1,
}


class OptimizerNames(ExplicitEnum):
    """
    Stores the acceptable string identifiers for optimizers.
    """

    ADAMW_HF = "adamw_hf"
    ADAMW_TORCH = "adamw_torch"
    ADAMW_TORCH_XLA = "adamw_torch_xla"
    ADAMW_APEX_FUSED = "adamw_apex_fused"
    ADAFACTOR = "adafactor"
    ADAMW_BNB = "adamw_bnb_8bit"
    SGD = "sgd"
    ADAGRAD = "adagrad"


@dataclass
class ClassWithEnum(TestSetup):
    evaluation_strategy: Union[IntervalStrategy, str] = field(
        default="no",
        metadata={"help": "The evaluation strategy to use."},
    )


def test_enums_are_parsed_to_enum_member():
    # NOTE: Since it's a union of str and IntervalStrategy, it shouldn't raise an error if given
    # an invalid value.
    assert ClassWithEnum.setup("--evaluation_strategy invalid").evaluation_strategy == "invalid"

    # However, it is, once we factor in what's happening in the __post_init__ of TrainingArguments.
    with pytest.raises(ValueError):
        TrainingArguments.setup("--evaluation_strategy invalid")

    for mode, enum_value in zip(
        ["no", "steps", "epoch"],
        [IntervalStrategy.NO, IntervalStrategy.STEPS, IntervalStrategy.EPOCH],
    ):
        assert (
            TrainingArguments.setup(f"--evaluation_strategy {mode}").evaluation_strategy
            == enum_value
        )


@dataclass
class TrainingArguments(TestSetup):
    """
    TrainingArguments is the subset of the arguments we use in our example scripts **which relate to the training loop
    itself**.
    Using [`HfArgumentParser`] we can turn this class into
    [argparse](https://docs.python.org/3/library/argparse#module-argparse) arguments that can be specified on the
    command line.
    Parameters:
        output_dir (`str`):
            The output directory where the model predictions and checkpoints will be written.
        overwrite_output_dir (`bool`, *optional*, defaults to `False`):
            If `True`, overwrite the content of the output directory. Use this to continue training if `output_dir`
            points to a checkpoint directory.
        do_train (`bool`, *optional*, defaults to `False`):
            Whether to run training or not. This argument is not directly used by [`Trainer`], it's intended to be used
            by your training/evaluation scripts instead. See the [example
            scripts](https://github.com/huggingface/transformers/tree/main/examples) for more details.
        do_eval (`bool`, *optional*):
            Whether to run evaluation on the validation set or not. Will be set to `True` if `evaluation_strategy` is
            different from `"no"`. This argument is not directly used by [`Trainer`], it's intended to be used by your
            training/evaluation scripts instead. See the [example
            scripts](https://github.com/huggingface/transformers/tree/main/examples) for more details.
        do_predict (`bool`, *optional*, defaults to `False`):
            Whether to run predictions on the test set or not. This argument is not directly used by [`Trainer`], it's
            intended to be used by your training/evaluation scripts instead. See the [example
            scripts](https://github.com/huggingface/transformers/tree/main/examples) for more details.
        evaluation_strategy (`str` or [`~trainer_utils.IntervalStrategy`], *optional*, defaults to `"no"`):
            The evaluation strategy to adopt during training. Possible values are:
                - `"no"`: No evaluation is done during training.
                - `"steps"`: Evaluation is done (and logged) every `eval_steps`.
                - `"epoch"`: Evaluation is done at the end of each epoch.
        prediction_loss_only (`bool`, *optional*, defaults to `False`):
            When performing evaluation and generating predictions, only returns the loss.
        per_device_train_batch_size (`int`, *optional*, defaults to 8):
            The batch size per GPU/TPU core/CPU for training.
        per_device_eval_batch_size (`int`, *optional*, defaults to 8):
            The batch size per GPU/TPU core/CPU for evaluation.
        gradient_accumulation_steps (`int`, *optional*, defaults to 1):
            Number of updates steps to accumulate the gradients for, before performing a backward/update pass.
            <Tip warning={true}>
            When using gradient accumulation, one step is counted as one step with backward pass. Therefore, logging,
            evaluation, save will be conducted every `gradient_accumulation_steps * xxx_step` training examples.
            </Tip>
        eval_accumulation_steps (`int`, *optional*):
            Number of predictions steps to accumulate the output tensors for, before moving the results to the CPU. If
            left unset, the whole predictions are accumulated on GPU/TPU before being moved to the CPU (faster but
            requires more memory).
        eval_delay (`float`, *optional*):
            Number of epochs or steps to wait for before the first evaluation can be performed, depending on the
            evaluation_strategy.
        learning_rate (`float`, *optional*, defaults to 5e-5):
            The initial learning rate for [`AdamW`] optimizer.
        weight_decay (`float`, *optional*, defaults to 0):
            The weight decay to apply (if not zero) to all layers except all bias and LayerNorm weights in [`AdamW`]
            optimizer.
        adam_beta1 (`float`, *optional*, defaults to 0.9):
            The beta1 hyperparameter for the [`AdamW`] optimizer.
        adam_beta2 (`float`, *optional*, defaults to 0.999):
            The beta2 hyperparameter for the [`AdamW`] optimizer.
        adam_epsilon (`float`, *optional*, defaults to 1e-8):
            The epsilon hyperparameter for the [`AdamW`] optimizer.
        max_grad_norm (`float`, *optional*, defaults to 1.0):
            Maximum gradient norm (for gradient clipping).
        num_train_epochs(`float`, *optional*, defaults to 3.0):
            Total number of training epochs to perform (if not an integer, will perform the decimal part percents of
            the last epoch before stopping training).
        max_steps (`int`, *optional*, defaults to -1):
            If set to a positive number, the total number of training steps to perform. Overrides `num_train_epochs`.
            In case of using a finite iterable dataset the training may stop before reaching the set number of steps
            when all data is exhausted
        lr_scheduler_type (`str` or [`SchedulerType`], *optional*, defaults to `"linear"`):
            The scheduler type to use. See the documentation of [`SchedulerType`] for all possible values.
        warmup_ratio (`float`, *optional*, defaults to 0.0):
            Ratio of total training steps used for a linear warmup from 0 to `learning_rate`.
        warmup_steps (`int`, *optional*, defaults to 0):
            Number of steps used for a linear warmup from 0 to `learning_rate`. Overrides any effect of `warmup_ratio`.
        log_level (`str`, *optional*, defaults to `passive`):
            Logger log level to use on the main process. Possible choices are the log levels as strings: 'debug',
            'info', 'warning', 'error' and 'critical', plus a 'passive' level which doesn't set anything and lets the
            application set the level.
        log_level_replica (`str`, *optional*, defaults to `passive`):
            Logger log level to use on replicas. Same choices as `log_level`"
        log_on_each_node (`bool`, *optional*, defaults to `True`):
            In multinode distributed training, whether to log using `log_level` once per node, or only on the main
            node.
        logging_dir (`str`, *optional*):
            [TensorBoard](https://www.tensorflow.org/tensorboard) log directory. Will default to
            *output_dir/runs/**CURRENT_DATETIME_HOSTNAME***.
        logging_strategy (`str` or [`~trainer_utils.IntervalStrategy`], *optional*, defaults to `"steps"`):
            The logging strategy to adopt during training. Possible values are:
                - `"no"`: No logging is done during training.
                - `"epoch"`: Logging is done at the end of each epoch.
                - `"steps"`: Logging is done every `logging_steps`.
        logging_first_step (`bool`, *optional*, defaults to `False`):
            Whether to log and evaluate the first `global_step` or not.
        logging_steps (`int`, *optional*, defaults to 500):
            Number of update steps between two logs if `logging_strategy="steps"`.
        logging_nan_inf_filter (`bool`, *optional*, defaults to `True`):
            Whether to filter `nan` and `inf` losses for logging. If set to `True` the loss of every step that is `nan`
            or `inf` is filtered and the average loss of the current logging window is taken instead.
            <Tip>
            `logging_nan_inf_filter` only influences the logging of loss values, it does not change the behavior the
            gradient is computed or applied to the model.
            </Tip>
        save_strategy (`str` or [`~trainer_utils.IntervalStrategy`], *optional*, defaults to `"steps"`):
            The checkpoint save strategy to adopt during training. Possible values are:
                - `"no"`: No save is done during training.
                - `"epoch"`: Save is done at the end of each epoch.
                - `"steps"`: Save is done every `save_steps`.
        save_steps (`int`, *optional*, defaults to 500):
            Number of updates steps before two checkpoint saves if `save_strategy="steps"`.
        save_total_limit (`int`, *optional*):
            If a value is passed, will limit the total amount of checkpoints. Deletes the older checkpoints in
            `output_dir`.
        save_on_each_node (`bool`, *optional*, defaults to `False`):
            When doing multi-node distributed training, whether to save models and checkpoints on each node, or only on
            the main one.
            This should not be activated when the different nodes use the same storage as the files will be saved with
            the same names for each node.
        no_cuda (`bool`, *optional*, defaults to `False`):
            Whether to not use CUDA even when it is available or not.
        seed (`int`, *optional*, defaults to 42):
            Random seed that will be set at the beginning of training. To ensure reproducibility across runs, use the
            [`~Trainer.model_init`] function to instantiate the model if it has some randomly initialized parameters.
        data_seed (`int`, *optional*):
            Random seed to be used with data samplers. If not set, random generators for data sampling will use the
            same seed as `seed`. This can be used to ensure reproducibility of data sampling, independent of the model
            seed.
        jit_mode_eval (`bool`, *optional*, defaults to `False`):
            Whether or not to use PyTorch jit trace for inference.
        use_ipex (`bool`, *optional*, defaults to `False`):
            Use Intel extension for PyTorch when it is available. [IPEX
            installation](https://github.com/intel/intel-extension-for-pytorch).
        bf16 (`bool`, *optional*, defaults to `False`):
            Whether to use bf16 16-bit (mixed) precision training instead of 32-bit training. Requires Ampere or higher
            NVIDIA architecture or using CPU (no_cuda). This is an experimental API and it may change.
        fp16 (`bool`, *optional*, defaults to `False`):
            Whether to use fp16 16-bit (mixed) precision training instead of 32-bit training.
        fp16_opt_level (`str`, *optional*, defaults to 'O1'):
            For `fp16` training, Apex AMP optimization level selected in ['O0', 'O1', 'O2', and 'O3']. See details on
            the [Apex documentation](https://nvidia.github.io/apex/amp).
        fp16_backend (`str`, *optional*, defaults to `"auto"`):
            This argument is deprecated. Use `half_precision_backend` instead.
        half_precision_backend (`str`, *optional*, defaults to `"auto"`):
            The backend to use for mixed precision training. Must be one of `"auto", "cuda_amp", "apex", "cpu_amp"`.
            `"auto"` will use CPU/CUDA AMP or APEX depending on the PyTorch version detected, while the other choices
            will force the requested backend.
        bf16_full_eval (`bool`, *optional*, defaults to `False`):
            Whether to use full bfloat16 evaluation instead of 32-bit. This will be faster and save memory but can harm
            metric values. This is an experimental API and it may change.
        fp16_full_eval (`bool`, *optional*, defaults to `False`):
            Whether to use full float16 evaluation instead of 32-bit. This will be faster and save memory but can harm
            metric values.
        tf32 (`bool`, *optional*):
            Whether to enable the TF32 mode, available in Ampere and newer GPU architectures. The default value depends
            on PyTorch's version default of `torch.backends.cuda.matmul.allow_tf32`. For more details please refer to
            the [TF32](https://huggingface.co/docs/transformers/performance#tf32) documentation. This is an
            experimental API and it may change.
        local_rank (`int`, *optional*, defaults to -1):
            Rank of the process during distributed training.
        xpu_backend (`str`, *optional*):
            The backend to use for xpu distributed training. Must be one of `"mpi"` or `"ccl"` or `"gloo"`.
        tpu_num_cores (`int`, *optional*):
            When training on TPU, the number of TPU cores (automatically passed by launcher script).
        dataloader_drop_last (`bool`, *optional*, defaults to `False`):
            Whether to drop the last incomplete batch (if the length of the dataset is not divisible by the batch size)
            or not.
        eval_steps (`int`, *optional*):
            Number of update steps between two evaluations if `evaluation_strategy="steps"`. Will default to the same
            value as `logging_steps` if not set.
        dataloader_num_workers (`int`, *optional*, defaults to 0):
            Number of subprocesses to use for data loading (PyTorch only). 0 means that the data will be loaded in the
            main process.
        past_index (`int`, *optional*, defaults to -1):
            Some models like [TransformerXL](../model_doc/transformerxl) or [XLNet](../model_doc/xlnet) can make use of
            the past hidden states for their predictions. If this argument is set to a positive int, the `Trainer` will
            use the corresponding output (usually index 2) as the past state and feed it to the model at the next
            training step under the keyword argument `mems`.
        run_name (`str`, *optional*):
            A descriptor for the run. Typically used for [wandb](https://www.wandb.com/) and
            [mlflow](https://www.mlflow.org/) logging.
        disable_tqdm (`bool`, *optional*):
            Whether or not to disable the tqdm progress bars and table of metrics produced by
            [`~notebook.NotebookTrainingTracker`] in Jupyter Notebooks. Will default to `True` if the logging level is
            set to warn or lower (default), `False` otherwise.
        remove_unused_columns (`bool`, *optional*, defaults to `True`):
            Whether or not to automatically remove the columns unused by the model forward method.
            (Note that this behavior is not implemented for [`TFTrainer`] yet.)
        label_names (`List[str]`, *optional*):
            The list of keys in your dictionary of inputs that correspond to the labels.
            Will eventually default to `["labels"]` except if the model used is one of the `XxxForQuestionAnswering` in
            which case it will default to `["start_positions", "end_positions"]`.
        load_best_model_at_end (`bool`, *optional*, defaults to `False`):
            Whether or not to load the best model found during training at the end of training.
            <Tip>
            When set to `True`, the parameters `save_strategy` needs to be the same as `evaluation_strategy`, and in
            the case it is "steps", `save_steps` must be a round multiple of `eval_steps`.
            </Tip>
        metric_for_best_model (`str`, *optional*):
            Use in conjunction with `load_best_model_at_end` to specify the metric to use to compare two different
            models. Must be the name of a metric returned by the evaluation with or without the prefix `"eval_"`. Will
            default to `"loss"` if unspecified and `load_best_model_at_end=True` (to use the evaluation loss).
            If you set this value, `greater_is_better` will default to `True`. Don't forget to set it to `False` if
            your metric is better when lower.
        greater_is_better (`bool`, *optional*):
            Use in conjunction with `load_best_model_at_end` and `metric_for_best_model` to specify if better models
            should have a greater metric or not. Will default to:
            - `True` if `metric_for_best_model` is set to a value that isn't `"loss"` or `"eval_loss"`.
            - `False` if `metric_for_best_model` is not set, or set to `"loss"` or `"eval_loss"`.
        ignore_data_skip (`bool`, *optional*, defaults to `False`):
            When resuming training, whether or not to skip the epochs and batches to get the data loading at the same
            stage as in the previous training. If set to `True`, the training will begin faster (as that skipping step
            can take a long time) but will not yield the same results as the interrupted training would have.
        sharded_ddp (`bool`, `str` or list of [`~trainer_utils.ShardedDDPOption`], *optional*, defaults to `False`):
            Use Sharded DDP training from [FairScale](https://github.com/facebookresearch/fairscale) (in distributed
            training only). This is an experimental feature.
            A list of options along the following:
            - `"simple"`: to use first instance of sharded DDP released by fairscale (`ShardedDDP`) similar to ZeRO-2.
            - `"zero_dp_2"`: to use the second instance of sharded DPP released by fairscale (`FullyShardedDDP`) in
              Zero-2 mode (with `reshard_after_forward=False`).
            - `"zero_dp_3"`: to use the second instance of sharded DPP released by fairscale (`FullyShardedDDP`) in
              Zero-3 mode (with `reshard_after_forward=True`).
            - `"offload"`: to add ZeRO-offload (only compatible with `"zero_dp_2"` and `"zero_dp_3"`).
            If a string is passed, it will be split on space. If a bool is passed, it will be converted to an empty
            list for `False` and `["simple"]` for `True`.
        fsdp (`bool`, `str` or list of [`~trainer_utils.FSDPOption`], *optional*, defaults to `False`):
            Use PyTorch Distributed Parallel Training (in distributed training only).
            A list of options along the following:
            - `"full_shard"`: Shard parameters, gradients and optimizer states.
            - `"shard_grad_op"`: Shard optimizer states and gradients.
            - `"offload"`: Offload parameters and gradients to CPUs (only compatible with `"full_shard"` and
              `"shard_grad_op"`).
            - `"auto_wrap"`: Automatically recursively wrap layers with FSDP using `default_auto_wrap_policy`.
        fsdp_min_num_params (`int`, *optional*, defaults to `0`):
            FSDP's minimum number of parameters for Default Auto Wrapping. (useful only when `fsdp` field is passed).
        deepspeed (`str` or `dict`, *optional*):
            Use [Deepspeed](https://github.com/microsoft/deepspeed). This is an experimental feature and its API may
            evolve in the future. The value is either the location of DeepSpeed json config file (e.g.,
            `ds_config.json`) or an already loaded json file as a `dict`"
        label_smoothing_factor (`float`, *optional*, defaults to 0.0):
            The label smoothing factor to use. Zero means no label smoothing, otherwise the underlying onehot-encoded
            labels are changed from 0s and 1s to `label_smoothing_factor/num_labels` and `1 - label_smoothing_factor +
            label_smoothing_factor/num_labels` respectively.
        debug (`str` or list of [`~debug_utils.DebugOption`], *optional*, defaults to `""`):
            Enable one or more debug features. This is an experimental feature.
            Possible options are:
            - `"underflow_overflow"`: detects overflow in model's input/outputs and reports the last frames that led to
              the event
            - `"tpu_metrics_debug"`: print debug metrics on TPU
            The options should be separated by whitespaces.
        optim (`str` or [`training_args.OptimizerNames`], *optional*, defaults to `"adamw_hf"`):
            The optimizer to use: adamw_hf, adamw_torch, adamw_apex_fused, or adafactor.
        adafactor (`bool`, *optional*, defaults to `False`):
            This argument is deprecated. Use `--optim adafactor` instead.
        group_by_length (`bool`, *optional*, defaults to `False`):
            Whether or not to group together samples of roughly the same length in the training dataset (to minimize
            padding applied and be more efficient). Only useful if applying dynamic padding.
        length_column_name (`str`, *optional*, defaults to `"length"`):
            Column name for precomputed lengths. If the column exists, grouping by length will use these values rather
            than computing them on train startup. Ignored unless `group_by_length` is `True` and the dataset is an
            instance of `Dataset`.
        report_to (`str` or `List[str]`, *optional*, defaults to `"all"`):
            The list of integrations to report the results and logs to. Supported platforms are `"azure_ml"`,
            `"comet_ml"`, `"mlflow"`, `"neptune"`, `"tensorboard"` and `"wandb"`. Use `"all"` to report to all
            integrations installed, `"none"` for no integrations.
        ddp_find_unused_parameters (`bool`, *optional*):
            When using distributed training, the value of the flag `find_unused_parameters` passed to
            `DistributedDataParallel`. Will default to `False` if gradient checkpointing is used, `True` otherwise.
        ddp_bucket_cap_mb (`int`, *optional*):
            When using distributed training, the value of the flag `bucket_cap_mb` passed to `DistributedDataParallel`.
        dataloader_pin_memory (`bool`, *optional*, defaults to `True`):
            Whether you want to pin memory in data loaders or not. Will default to `True`.
        skip_memory_metrics (`bool`, *optional*, defaults to `True`):
            Whether to skip adding of memory profiler reports to metrics. This is skipped by default because it slows
            down the training and evaluation speed.
        push_to_hub (`bool`, *optional*, defaults to `False`):
            Whether or not to push the model to the Hub every time the model is saved. If this is activated,
            `output_dir` will begin a git directory synced with the repo (determined by `hub_model_id`) and the content
            will be pushed each time a save is triggered (depending on your `save_strategy`). Calling
            [`~Trainer.save_model`] will also trigger a push.
            <Tip warning={true}>
            If `output_dir` exists, it needs to be a local clone of the repository to which the [`Trainer`] will be
            pushed.
            </Tip>
        resume_from_checkpoint (`str`, *optional*):
            The path to a folder with a valid checkpoint for your model. This argument is not directly used by
            [`Trainer`], it's intended to be used by your training/evaluation scripts instead. See the [example
            scripts](https://github.com/huggingface/transformers/tree/main/examples) for more details.
        hub_model_id (`str`, *optional*):
            The name of the repository to keep in sync with the local *output_dir*. It can be a simple model ID in
            which case the model will be pushed in your namespace. Otherwise it should be the whole repository name,
            for instance `"user_name/model"`, which allows you to push to an organization you are a member of with
            `"organization_name/model"`. Will default to `user_name/output_dir_name` with *output_dir_name* being the
            name of `output_dir`.
            Will default to the name of `output_dir`.
        hub_strategy (`str` or [`~trainer_utils.HubStrategy`], *optional*, defaults to `"every_save"`):
            Defines the scope of what is pushed to the Hub and when. Possible values are:
            - `"end"`: push the model, its configuration, the tokenizer (if passed along to the [`Trainer`]) and a
              draft of a model card when the [`~Trainer.save_model`] method is called.
            - `"every_save"`: push the model, its configuration, the tokenizer (if passed along to the [`Trainer`]) and
              a draft of a model card each time there is a model save. The pushes are asynchronous to not block
              training, and in case the save are very frequent, a new push is only attempted if the previous one is
              finished. A last push is made with the final model at the end of training.
            - `"checkpoint"`: like `"every_save"` but the latest checkpoint is also pushed in a subfolder named
              last-checkpoint, allowing you to resume training easily with
              `trainer.train(resume_from_checkpoint="last-checkpoint")`.
            - `"all_checkpoints"`: like `"checkpoint"` but all checkpoints are pushed like they appear in the output
              folder (so you will get one checkpoint folder per folder in your final repository)
        hub_token (`str`, *optional*):
            The token to use to push the model to the Hub. Will default to the token in the cache folder obtained with
            `huggingface-cli login`.
        hub_private_repo (`bool`, *optional*, defaults to `False`):
            If True, the Hub repo will be set to private.
        gradient_checkpointing (`bool`, *optional*, defaults to `False`):
            If True, use gradient checkpointing to save memory at the expense of slower backward pass.
        include_inputs_for_metrics (`bool`, *optional*, defaults to `False`):
            Whether or not the inputs will be passed to the `compute_metrics` function. This is intended for metrics
            that need inputs, predictions and references for scoring calculation in Metric class.
        auto_find_batch_size (`bool`, *optional*, defaults to `False`)
            Whether to find a batch size that will fit into memory automatically through exponential decay, avoiding
            CUDA Out-of-Memory errors. Requires accelerate to be installed (`pip install accelerate`)
        full_determinism (`bool`, *optional*, defaults to `False`)
            If `True`, [`enable_full_determinism`] is called instead of [`set_seed`] to ensure reproducible results in
            distributed training
        torchdynamo (`str`, *optional*):
            The token that is used to set the backend compiler for TorchDynamo. Possible choices are ["eager",
            "nvfuser]. This is an experimental API and subject to change.
        ray_scope (`str`, *optional*, defaults to `"last"`):
            The scope to use when doing hyperparameter search with Ray. By default, `"last"` will be used. Ray will
            then use the last checkpoint of all trials, compare those, and select the best one. However, other options
            are also available. See the [Ray documentation](
            https://docs.ray.io/en/latest/tune/api_docs/analysis.html#ray.tune.ExperimentAnalysis.get_best_trial) for
            more options.
        ddp_timeout (`int`, *optional*, defaults to 1800):
            The timeout for `torch.distributed.init_process_group` calls, used to avoid GPU socket timeouts when
            performing slow operations in distributed runnings. Please refer the [PyTorch documentation]
            (https://pytorch.org/docs/stable/distributed.html#torch.distributed.init_process_group) for more
            information.
        use_mps_device (`bool`, *optional*, defaults to `False`):
            Whether to use Apple Silicon chip based `mps` device.
    """

    framework = "pt"
    output_dir: str = field(
        default="tmp_trainer",
        metadata={
            "help": "The output directory where the model predictions and checkpoints will be written."
        },
    )
    overwrite_output_dir: bool = field(
        default=False,
        metadata={
            "help": (
                "Overwrite the content of the output directory. "
                "Use this to continue training if output_dir points to a checkpoint directory."
            )
        },
    )

    do_train: bool = field(default=False, metadata={"help": "Whether to run training."})
    do_eval: bool = field(default=False, metadata={"help": "Whether to run eval on the dev set."})
    do_predict: bool = field(
        default=False, metadata={"help": "Whether to run predictions on the test set."}
    )
    evaluation_strategy: Union[IntervalStrategy, str] = field(
        default="no",
        metadata={"help": "The evaluation strategy to use."},
    )
    prediction_loss_only: bool = field(
        default=False,
        metadata={"help": "When performing evaluation and predictions, only returns the loss."},
    )

    per_device_train_batch_size: int = field(
        default=8, metadata={"help": "Batch size per GPU/TPU core/CPU for training."}
    )
    per_device_eval_batch_size: int = field(
        default=8, metadata={"help": "Batch size per GPU/TPU core/CPU for evaluation."}
    )

    per_gpu_train_batch_size: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "Deprecated, the use of `--per_device_train_batch_size` is preferred. "
                "Batch size per GPU/TPU core/CPU for training."
            )
        },
    )
    per_gpu_eval_batch_size: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "Deprecated, the use of `--per_device_eval_batch_size` is preferred. "
                "Batch size per GPU/TPU core/CPU for evaluation."
            )
        },
    )

    gradient_accumulation_steps: int = field(
        default=1,
        metadata={
            "help": "Number of updates steps to accumulate before performing a backward/update pass."
        },
    )
    eval_accumulation_steps: Optional[int] = field(
        default=None,
        metadata={
            "help": "Number of predictions steps to accumulate before moving the tensors to the CPU."
        },
    )

    eval_delay: Optional[float] = field(
        default=0,
        metadata={
            "help": (
                "Number of epochs or steps to wait for before the first evaluation can be performed, depending on the"
                " evaluation_strategy."
            )
        },
    )

    learning_rate: float = field(
        default=5e-5, metadata={"help": "The initial learning rate for AdamW."}
    )
    weight_decay: float = field(
        default=0.0, metadata={"help": "Weight decay for AdamW if we apply some."}
    )
    adam_beta1: float = field(default=0.9, metadata={"help": "Beta1 for AdamW optimizer"})
    adam_beta2: float = field(default=0.999, metadata={"help": "Beta2 for AdamW optimizer"})
    adam_epsilon: float = field(default=1e-8, metadata={"help": "Epsilon for AdamW optimizer."})
    max_grad_norm: float = field(default=1.0, metadata={"help": "Max gradient norm."})

    num_train_epochs: float = field(
        default=3.0, metadata={"help": "Total number of training epochs to perform."}
    )
    max_steps: int = field(
        default=-1,
        metadata={
            "help": "If > 0: set total number of training steps to perform. Override num_train_epochs."
        },
    )
    lr_scheduler_type: Union[SchedulerType, str] = field(
        default="linear",
        metadata={"help": "The scheduler type to use."},
    )
    warmup_ratio: float = field(
        default=0.0, metadata={"help": "Linear warmup over warmup_ratio fraction of total steps."}
    )
    warmup_steps: int = field(default=0, metadata={"help": "Linear warmup over warmup_steps."})

    log_level: Optional[str] = field(
        default="passive",
        metadata={
            "help": (
                "Logger log level to use on the main node. Possible choices are the log levels as strings: 'debug',"
                " 'info', 'warning', 'error' and 'critical', plus a 'passive' level which doesn't set anything and"
                " lets the application set the level. Defaults to 'passive'."
            ),
            "choices": trainer_log_levels.keys(),
        },
    )
    log_level_replica: Optional[str] = field(
        default="passive",
        metadata={
            "help": "Logger log level to use on replica nodes. Same choices and defaults as ``log_level``",
            "choices": trainer_log_levels.keys(),
        },
    )
    log_on_each_node: bool = field(
        default=True,
        metadata={
            "help": (
                "When doing a multinode distributed training, whether to log once per node or just once on the main"
                " node."
            )
        },
    )
    logging_dir: Optional[str] = field(default=None, metadata={"help": "Tensorboard log dir."})
    logging_strategy: Union[IntervalStrategy, str] = field(
        default="steps",
        metadata={"help": "The logging strategy to use."},
    )
    logging_first_step: bool = field(default=False, metadata={"help": "Log the first global_step"})
    logging_steps: int = field(default=500, metadata={"help": "Log every X updates steps."})
    logging_nan_inf_filter: bool = field(
        default=True, metadata={"help": "Filter nan and inf losses for logging."}
    )
    save_strategy: Union[IntervalStrategy, str] = field(
        default="steps",
        metadata={"help": "The checkpoint save strategy to use."},
    )
    save_steps: int = field(
        default=500, metadata={"help": "Save checkpoint every X updates steps."}
    )
    save_total_limit: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "Limit the total amount of checkpoints. "
                "Deletes the older checkpoints in the output_dir. Default is unlimited checkpoints"
            )
        },
    )
    save_on_each_node: bool = field(
        default=False,
        metadata={
            "help": (
                "When doing multi-node distributed training, whether to save models and checkpoints on each node, or"
                " only on the main one"
            )
        },
    )
    no_cuda: bool = field(
        default=False, metadata={"help": "Do not use CUDA even when it is available"}
    )
    use_mps_device: bool = field(
        default=False, metadata={"help": "Whether to use Apple Silicon chip based `mps` device."}
    )
    seed: int = field(
        default=42, metadata={"help": "Random seed that will be set at the beginning of training."}
    )
    data_seed: Optional[int] = field(
        default=None, metadata={"help": "Random seed to be used with data samplers."}
    )
    jit_mode_eval: bool = field(
        default=False, metadata={"help": "Whether or not to use PyTorch jit trace for inference"}
    )
    use_ipex: bool = field(
        default=False,
        metadata={
            "help": (
                "Use Intel extension for PyTorch when it is available, installation:"
                " 'https://github.com/intel/intel-extension-for-pytorch'"
            )
        },
    )
    bf16: bool = field(
        default=False,
        metadata={
            "help": (
                "Whether to use bf16 (mixed) precision instead of 32-bit. Requires Ampere or higher NVIDIA"
                " architecture or using CPU (no_cuda). This is an experimental API and it may change."
            )
        },
    )
    fp16: bool = field(
        default=False,
        metadata={"help": "Whether to use fp16 (mixed) precision instead of 32-bit"},
    )
    fp16_opt_level: str = field(
        default="O1",
        metadata={
            "help": (
                "For fp16: Apex AMP optimization level selected in ['O0', 'O1', 'O2', and 'O3']. "
                "See details at https://nvidia.github.io/apex/amp.html"
            )
        },
    )
    half_precision_backend: str = field(
        default="auto",
        metadata={
            "help": "The backend to be used for half precision.",
            "choices": ["auto", "cuda_amp", "apex", "cpu_amp"],
        },
    )
    bf16_full_eval: bool = field(
        default=False,
        metadata={
            "help": (
                "Whether to use full bfloat16 evaluation instead of 32-bit. This is an experimental API and it may"
                " change."
            )
        },
    )
    fp16_full_eval: bool = field(
        default=False,
        metadata={"help": "Whether to use full float16 evaluation instead of 32-bit"},
    )
    tf32: Optional[bool] = field(
        default=None,
        metadata={
            "help": (
                "Whether to enable tf32 mode, available in Ampere and newer GPU architectures. This is an experimental"
                " API and it may change."
            )
        },
    )
    local_rank: int = field(default=-1, metadata={"help": "For distributed training: local_rank"})
    xpu_backend: Optional[str] = field(
        default=None,
        metadata={
            "help": "The backend to be used for distributed training on Intel XPU.",
            "choices": ["mpi", "ccl", "gloo"],
        },
    )
    tpu_num_cores: Optional[int] = field(
        default=None,
        metadata={"help": "TPU: Number of TPU cores (automatically passed by launcher script)"},
    )
    tpu_metrics_debug: bool = field(
        default=False,
        metadata={
            "help": (
                "Deprecated, the use of `--debug tpu_metrics_debug` is preferred. TPU: Whether to print debug metrics"
            )
        },
    )
    debug: str = field(
        default="",
        metadata={
            "help": (
                "Whether or not to enable debug mode. Current options: "
                "`underflow_overflow` (Detect underflow and overflow in activations and weights), "
                "`tpu_metrics_debug` (print debug metrics on TPU)."
            )
        },
    )

    dataloader_drop_last: bool = field(
        default=False,
        metadata={
            "help": "Drop the last incomplete batch if it is not divisible by the batch size."
        },
    )
    eval_steps: Optional[int] = field(
        default=None, metadata={"help": "Run an evaluation every X steps."}
    )
    dataloader_num_workers: int = field(
        default=0,
        metadata={
            "help": (
                "Number of subprocesses to use for data loading (PyTorch only). 0 means that the data will be loaded"
                " in the main process."
            )
        },
    )

    past_index: int = field(
        default=-1,
        metadata={
            "help": "If >=0, uses the corresponding part of the output as the past state for next step."
        },
    )

    run_name: Optional[str] = field(
        default=None,
        metadata={"help": "An optional descriptor for the run. Notably used for wandb logging."},
    )
    disable_tqdm: Optional[bool] = field(
        default=None, metadata={"help": "Whether or not to disable the tqdm progress bars."}
    )

    remove_unused_columns: Optional[bool] = field(
        default=True,
        metadata={"help": "Remove columns not required by the model when using an nlp.Dataset."},
    )
    label_names: Optional[List[str]] = field(
        default=None,
        metadata={
            "help": "The list of keys in your dictionary of inputs that correspond to the labels."
        },
    )

    load_best_model_at_end: Optional[bool] = field(
        default=False,
        metadata={
            "help": "Whether or not to load the best model found during training at the end of training."
        },
    )
    metric_for_best_model: Optional[str] = field(
        default=None, metadata={"help": "The metric to use to compare two different models."}
    )
    greater_is_better: Optional[bool] = field(
        default=None,
        metadata={"help": "Whether the `metric_for_best_model` should be maximized or not."},
    )
    ignore_data_skip: bool = field(
        default=False,
        metadata={
            "help": (
                "When resuming training, whether or not to skip the first epochs and batches to get to the same"
                " training data."
            )
        },
    )
    sharded_ddp: str = field(
        default="",
        metadata={
            "help": (
                "Whether or not to use sharded DDP training (in distributed training only). The base option should be"
                " `simple`, `zero_dp_2` or `zero_dp_3` and you can add CPU-offload to `zero_dp_2` or `zero_dp_3` like"
                " this: zero_dp_2 offload` or `zero_dp_3 offload`. You can add auto-wrap to `zero_dp_2` or `zero_dp_3`"
                " with the same syntax: zero_dp_2 auto_wrap` or `zero_dp_3 auto_wrap`."
            ),
        },
    )
    fsdp: str = field(
        default="",
        metadata={
            "help": (
                "Whether or not to use PyTorch Fully Sharded Data Parallel (FSDP) training (in distributed training"
                " only). The base option should be `full_shard`, `shard_grad_op` or `no_shard` and you can add"
                " CPU-offload to `full_shard` or `shard_grad_op` like this: full_shard offload` or `shard_grad_op"
                " offload`. You can add auto-wrap to `full_shard` or `shard_grad_op` with the same syntax: full_shard"
                " auto_wrap` or `shard_grad_op auto_wrap`."
            ),
        },
    )
    fsdp_min_num_params: int = field(
        default=0,
        metadata={
            "help": (
                "FSDP's minimum number of parameters for Default Auto Wrapping. (useful only when `fsdp` field is"
                " passed)."
            )
        },
    )
    fsdp_transformer_layer_cls_to_wrap: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "Transformer layer class name (case-sensitive) to wrap ,e.g, `BertLayer`, `GPTJBlock`, `T5Block` .... "
                "(useful only when `fsdp` flag is passed)."
            )
        },
    )
    deepspeed: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "Enable deepspeed and pass the path to deepspeed json config file (e.g. ds_config.json) or an already"
                " loaded json file as a dict"
            )
        },
    )
    label_smoothing_factor: float = field(
        default=0.0,
        metadata={"help": "The label smoothing epsilon to apply (zero means no label smoothing)."},
    )
    optim: Union[OptimizerNames, str] = field(
        default="adamw_hf",
        metadata={"help": "The optimizer to use."},
    )
    adafactor: bool = field(
        default=False, metadata={"help": "Whether or not to replace AdamW by Adafactor."}
    )
    group_by_length: bool = field(
        default=False,
        metadata={
            "help": "Whether or not to group samples of roughly the same length together when batching."
        },
    )
    length_column_name: Optional[str] = field(
        default="length",
        metadata={"help": "Column name with precomputed lengths to use when grouping by length."},
    )
    report_to: Optional[List[str]] = field(
        default=None,
        metadata={"help": "The list of integrations to report the results and logs to."},
    )
    ddp_find_unused_parameters: Optional[bool] = field(
        default=None,
        metadata={
            "help": (
                "When using distributed training, the value of the flag `find_unused_parameters` passed to "
                "`DistributedDataParallel`."
            )
        },
    )
    ddp_bucket_cap_mb: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "When using distributed training, the value of the flag `bucket_cap_mb` passed to "
                "`DistributedDataParallel`."
            )
        },
    )
    dataloader_pin_memory: bool = field(
        default=True, metadata={"help": "Whether or not to pin memory for DataLoader."}
    )
    skip_memory_metrics: bool = field(
        default=True,
        metadata={"help": "Whether or not to skip adding of memory profiler reports to metrics."},
    )
    use_legacy_prediction_loop: bool = field(
        default=False,
        metadata={"help": "Whether or not to use the legacy prediction_loop in the Trainer."},
    )
    push_to_hub: bool = field(
        default=False,
        metadata={
            "help": "Whether or not to upload the trained model to the model hub after training."
        },
    )
    resume_from_checkpoint: Optional[str] = field(
        default=None,
        metadata={"help": "The path to a folder with a valid checkpoint for your model."},
    )
    hub_model_id: Optional[str] = field(
        default=None,
        metadata={
            "help": "The name of the repository to keep in sync with the local `output_dir`."
        },
    )
    hub_strategy: Union[HubStrategy, str] = field(
        default="every_save",
        metadata={"help": "The hub strategy to use when `--push_to_hub` is activated."},
    )
    hub_token: Optional[str] = field(
        default=None, metadata={"help": "The token to use to push to the Model Hub."}
    )
    hub_private_repo: bool = field(
        default=False, metadata={"help": "Whether the model repository is private or not."}
    )
    gradient_checkpointing: bool = field(
        default=False,
        metadata={
            "help": "If True, use gradient checkpointing to save memory at the expense of slower backward pass."
        },
    )
    include_inputs_for_metrics: bool = field(
        default=False,
        metadata={
            "help": "Whether or not the inputs will be passed to the `compute_metrics` function."
        },
    )
    # Deprecated arguments
    fp16_backend: str = field(
        default="auto",
        metadata={
            "help": "Deprecated. Use half_precision_backend instead",
            "choices": ["auto", "cuda_amp", "apex", "cpu_amp"],
        },
    )
    push_to_hub_model_id: Optional[str] = field(
        default=None, metadata={"help": "The name of the repository to which push the `Trainer`."}
    )
    push_to_hub_organization: Optional[str] = field(
        default=None,
        metadata={"help": "The name of the organization in with to which push the `Trainer`."},
    )
    push_to_hub_token: Optional[str] = field(
        default=None, metadata={"help": "The token to use to push to the Model Hub."}
    )
    _n_gpu: int = field(init=False, repr=False, default=-1)
    mp_parameters: str = field(
        default="",
        metadata={
            "help": "Used by the SageMaker launcher to send mp-specific args. Ignored in Trainer"
        },
    )

    auto_find_batch_size: bool = field(
        default=False,
        metadata={
            "help": (
                "Whether to automatically decrease the batch size in half and rerun the training loop again each time"
                " a CUDA Out-of-Memory was reached"
            )
        },
    )
    full_determinism: bool = field(
        default=False,
        metadata={
            "help": (
                "Whether to call enable_full_determinism instead of set_seed for reproducibility in distributed"
                " training"
            )
        },
    )
    torchdynamo: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "Sets up the backend compiler for TorchDynamo. TorchDynamo is a Python level JIT compiler designed to"
                " make unmodified PyTorch programs faster. TorchDynamo dynamically modifies the Python bytecode right"
                " before its executed. It rewrites Python bytecode to extract sequences of PyTorch operations"
                " and lifts them up into Fx graph. We can then pass these Fx graphs to other backend compilers. There"
                " are two options - eager and nvfuser. Eager defaults to pytorch eager and is useful for debugging."
                " nvfuser path uses AOT Autograd and nvfuser compiler to optimize the models."
            ),
            "choices": ["eager", "nvfuser", "fx2trt", "fx2trt-fp16"],
        },
    )
    ray_scope: Optional[str] = field(
        default="last",
        metadata={
            "help": (
                'The scope to use when doing hyperparameter search with Ray. By default, `"last"` will be used. Ray'
                " will then use the last checkpoint of all trials, compare those, and select the best one. However,"
                " other options are also available. See the Ray documentation"
                " (https://docs.ray.io/en/latest/tune/api_docs/analysis.html"
                "#ray.tune.ExperimentAnalysis.get_best_trial)"
                " for more options."
            )
        },
    )
    ddp_timeout: Optional[int] = field(
        default=1800,
        metadata={
            "help": "Overrides the default timeout for distributed training (value should be given in seconds)."
        },
    )

    def __post_init__(self):
        # NOTE: Just a portion of the actual __post_init__ of the TrainingArguments class which is
        # causing issues.
        import os

        # Handle --use_env option in torch.distributed.launch (local_rank not passed as an arg then).
        # This needs to happen before any call to self.device or self.n_gpu.
        env_local_rank = int(os.environ.get("LOCAL_RANK", -1))
        if env_local_rank != -1 and env_local_rank != self.local_rank:
            self.local_rank = env_local_rank

        # expand paths, if not os.makedirs("~/bar") will make directory
        # in the current directory instead of the actual home
        #  see https://github.com/huggingface/transformers/issues/10628
        if self.output_dir is not None:
            self.output_dir = os.path.expanduser(self.output_dir)
        if self.logging_dir is None and self.output_dir is not None:
            self.logging_dir = os.path.join(self.output_dir, "tmp_trainer")
        if self.logging_dir is not None:
            self.logging_dir = os.path.expanduser(self.logging_dir)

        # if self.disable_tqdm is None:
        #     self.disable_tqdm = logger.getEffectiveLevel() > logging.WARN
        import warnings

        if isinstance(self.evaluation_strategy, EvaluationStrategy):
            warnings.warn(
                "using `EvaluationStrategy` for `evaluation_strategy` is deprecated and will be removed in version 5"
                " of 🤗 Transformers. Use `IntervalStrategy` instead",
                FutureWarning,
            )
            # Go back to the underlying string or we won't be able to instantiate `IntervalStrategy` on it.
            self.evaluation_strategy = self.evaluation_strategy.value

        self.evaluation_strategy = IntervalStrategy(self.evaluation_strategy)
        self.logging_strategy = IntervalStrategy(self.logging_strategy)
        self.save_strategy = IntervalStrategy(self.save_strategy)
        self.hub_strategy = HubStrategy(self.hub_strategy)


@pytest.mark.xfail(reason="docstring_parser can't parse the docstring of TrainingArguments!")
def test_docstring_parse_works_with_hf_training_args():

    assert get_attribute_docstring(TrainingArguments, "output_dir").desc_from_cls_docstring == (
        "The output directory where the model predictions and checkpoints will be written."
    )


def test_entire_docstring_isnt_used_as_help():
    help_text = TrainingArguments.get_help_text()

    help_from_field = "Whether to use Apple Silicon chip based `mps` device."
    assert help_from_field in help_text
    help_from_docstring = "use_mps_device (`bool`, *optional*, defaults to `False`):"
    assert help_from_docstring not in help_text


@pytest.mark.parametrize(
    "args",
    [
        TrainingArguments(),
        TrainingArguments(save_strategy="steps"),
        TrainingArguments(save_strategy=IntervalStrategy.EPOCH),
    ],
)
@pytest.mark.parametrize("filename", ["bob.yaml", "bob.json", "bob.pkl", "bob.yml"])
def test_serialization(tmp_path: Path, filename: str, args: TrainingArguments):
    """test that serializing / deserializing a TrainingArguments works."""
    from simple_parsing.helpers.serialization import load, save

    path = tmp_path / filename
    save(args, path)
    assert load(TrainingArguments, path) == args



import collections
import inspect
import sys
import types
import typing
from contextlib import contextmanager
from dataclasses import InitVar
from itertools import dropwhile
from logging import getLogger as get_logger
from typing import Any, Dict, Iterator, Optional, get_type_hints

logger = get_logger(__name__)

# NOTE: This dict is used to enable forward compatibility with things such as `tuple[int, str]`,
# `list[float]`, etc. when using `from __future__ import annotations`.
forward_refs_to_types = {
    "tuple": typing.Tuple,
    "set": typing.Set,
    "dict": typing.Dict,
    "list": typing.List,
    "type": typing.Type,
}


@contextmanager
def _initvar_patcher() -> Iterator[None]:
    """
    Patch InitVar to not fail when annotations are postponed.

    `TypeVar('Forward references must evaluate to types. Got dataclasses.InitVar[tp].')` is raised
    when postponed annotations are enabled and `get_type_hints` is called
    Bug is mentioned here https://github.com/python/cpython/issues/88962
    In python 3.11 this is fixed, but backport fix is not planned for old releases

    Workaround is mentioned here https://stackoverflow.com/q/70400639
    """
    if sys.version_info[:2] < (3, 11):
        InitVar.__call__ = lambda *args: None
    yield
    if sys.version_info[:2] < (3, 11):
        del InitVar.__call__


def evaluate_string_annotation(annotation: str, containing_class: Optional[type] = None) -> type:
    """Attempts to evaluate the given annotation string, to get a 'live' type annotation back.

    Any exceptions that are raised when evaluating are raised directly as-is.

    NOTE: This is probably not 100% safe. I mean, if the user code puts urls and stuff in their
    type annotations, and then uses simple-parsing, then sure, that code might get executed. But
    I don't think it's my job to prevent them from shooting themselves in the foot, you know what I
    mean?
    """
    # The type of the field might be a string when using `from __future__ import annotations`.
    # Get the local and global namespaces to pass to the `get_type_hints` function.
    local_ns: Dict[str, Any] = {"typing": typing, **vars(typing)}
    local_ns.update(forward_refs_to_types)
    global_ns = {}
    if containing_class:
        # Get the globals in the module where the class was defined.
        global_ns = sys.modules[containing_class.__module__].__dict__

    if "|" in annotation:
        annotation = _get_old_style_annotation(annotation)
    evaluated_t: type = eval(annotation, local_ns, global_ns)
    return evaluated_t


def _replace_UnionType_with_typing_Union(annotation):
    from simple_parsing.utils import builtin_types, is_dict, is_list, is_tuple

    if sys.version_info[:2] < (3, 10):
        # This is only useful for python 3.10+ (where UnionTypes exist).
        # Therefore just return the annotation as-is.
        return annotation

    if isinstance(annotation, types.UnionType):  # type: ignore
        union_args = typing.get_args(annotation)
        new_union_args = tuple(_replace_UnionType_with_typing_Union(arg) for arg in union_args)
        return typing.Union[new_union_args]  # type: ignore
    if is_list(annotation):
        item_annotation = typing.get_args(annotation)[0]
        new_item_annotation = _replace_UnionType_with_typing_Union(item_annotation)
        return typing.List[new_item_annotation]
    if is_tuple(annotation):
        item_annotations = typing.get_args(annotation)
        new_item_annotations = tuple(
            _replace_UnionType_with_typing_Union(arg) for arg in item_annotations
        )
        return typing.Tuple[new_item_annotations]  # type: ignore
    if is_dict(annotation):
        annotations = typing.get_args(annotation)
        if not annotations:
            return typing.Dict
        assert len(annotations) == 2
        key_annotation = annotations[0]
        value_annotation = annotations[1]
        new_key_annotation = _replace_UnionType_with_typing_Union(key_annotation)
        new_value_annotation = _replace_UnionType_with_typing_Union(value_annotation)
        return typing.Dict[new_key_annotation, new_value_annotation]
    if annotation in builtin_types:
        return annotation
    if inspect.isclass(annotation):
        return annotation
    raise NotImplementedError(annotation)


#     # return forward_refs_to_types.get(ann, local_ns.get(ann, global_ns.get(ann, getattr(builtins, ann, ann))))


def _not_supported(annotation) -> typing.NoReturn:
    raise NotImplementedError(f"Don't yet support annotations like this: {annotation}")


def _get_old_style_annotation(annotation: str) -> str:
    """Replaces A | B with Union[A,B] in the annotation."""
    # TODO: Add proper support for things like `list[int | float]`, which isn't currently
    # working, even without the new-style union.
    if "|" not in annotation:
        return annotation

    annotation = annotation.strip()
    if "[" not in annotation:
        assert "]" not in annotation
        return "Union[" + ", ".join(v.strip() for v in annotation.split("|")) + "]"

    before, lsep, rest = annotation.partition("[")
    middle, rsep, after = rest.rpartition("]")
    # BUG: Need to handle things like bob[int] | None
    assert (
        not after.strip()
    ), f"can't have text at HERE in <something>[<something>]<HERE>!: {annotation}"

    if "|" in before or "|" in after:
        _not_supported(annotation)
    assert "|" in middle

    if "," in middle:
        parts = [v.strip() for v in middle.split(",")]
        parts = [_get_old_style_annotation(part) for part in parts]
        middle = ", ".join(parts)

    new_middle = _get_old_style_annotation(annotation=middle)
    new_annotation = before + lsep + new_middle + rsep + after
    return new_annotation


def _replace_new_union_syntax_with_old_union_syntax(
    annotations_dict: Dict[str, str], context: collections.ChainMap
) -> Dict[str, Any]:

    new_annotations = annotations_dict.copy()
    for field, annotation_str in annotations_dict.items():
        updated_annotation = _get_old_style_annotation(annotation_str)
        new_annotations[field] = updated_annotation

    return new_annotations


def get_field_type_from_annotations(some_class: type, field_name: str) -> type:
    """Get the annotation for the given field, in the 'old-style' format with types from
    typing.List, typing.Union, etc.

    If the script uses `from __future__ import annotations`, and we are in python<3.9,
    Then we need to actually first make this forward-compatibility 'patch' so that we
    don't run into a "`type` object is not subscriptable" error.

    NOTE: If you get errors of this kind from the function below, then you might want to add an
    entry to the `forward_refs_to_types` dict above.
    """

    # Pretty hacky: Modify the type annotations of the class (preferably a copy of the class
    # if possible, to avoid modifying things in-place), and replace  the `a | b`-type
    # expressions with `Union[a, b]`, so that `get_type_hints` doesn't raise an error.
    # The type of the field might be a string when using `from __future__ import annotations`.

    # The type of the field might be a string when using `from __future__ import annotations`.
    # Get the local and global namespaces to pass to the `get_type_hints` function.
    local_ns: Dict[str, Any] = {"typing": typing, **vars(typing)}
    local_ns.update(forward_refs_to_types)

    # NOTE: Get the local namespace of the calling function / module where this class is defined,
    # and use it to get the correct type of the field, if it is a forward reference.
    frame = inspect.currentframe()
    # stack = []
    while frame.f_back is not None and frame.f_locals.get(some_class.__name__) is not some_class:
        # stack.append(frame)
        frame = frame.f_back
    # Found the frame with the dataclass definition. Update the locals. This makes it possible to
    # use dataclasses defined in local scopes!
    if frame is not None:
        local_ns.update(frame.f_locals)

    # Get the global_ns in the module starting from the deepest base until the module with the field_name last definition.
    global_ns = {}
    classes_to_iterate = list(dropwhile(
        lambda cls: field_name not in getattr(cls, "__annotations__", {}),
        some_class.mro()
    ))
    for base_cls in reversed(classes_to_iterate):
        global_ns.update(sys.modules[base_cls.__module__].__dict__)

    try:
        with _initvar_patcher():
            annotations_dict = get_type_hints(some_class, localns=local_ns, globalns=global_ns)
    except TypeError:
        annotations_dict = collections.ChainMap(
            *[getattr(cls, "__annotations__", {}) for cls in some_class.mro()]
        )

    if field_name not in annotations_dict:
        raise ValueError(f"Field {field_name} not found in annotations of class {some_class}")

    field_type = annotations_dict[field_name]

    if sys.version_info[:2] >= (3, 7) and isinstance(field_type, typing.ForwardRef):
        # Weird bug happens when mixing postponed evaluation of type annotations + forward
        # references: The ForwardRefs are left as-is, and not evaluated!
        forward_arg = field_type.__forward_arg__
        field_type = forward_arg

    if sys.version_info >= (3, 10) and isinstance(field_type, types.UnionType):
        # In python >= 3.10, int | float is allowed. Therefore, just to be consistent, we want
        # to convert those into the corresponding typing.Union type.
        # This is necessary for the rest of the code to work, since it's all based on typing.Union.
        field_type = _replace_UnionType_with_typing_Union(field_type)

    if isinstance(field_type, str) and "|" in field_type:
        field_type = _get_old_style_annotation(field_type)

        # Pretty hacky:
        # In order to use `get_type_hints`, we need to pass it a class. We can't just ask it to
        # evaluate a single annotation. Therefore, we create a temporary class and set it's
        # __annotation__ attribute, which is introspected by `get_type_hints`.

    try:

        class Temp_:
            pass

        Temp_.__annotations__ = {field_name: field_type}
        with _initvar_patcher():
            annotations_dict = get_type_hints(Temp_, globalns=global_ns, localns=local_ns)
        field_type = annotations_dict[field_name]
    except Exception:
        logger.warning(
            f"Unable to evaluate forward reference {field_type} for field '{field_name}'.\n"
            f"Leaving it as-is."
        )
        field_type = field_type

    return field_type



import textwrap
from dataclasses import dataclass
from io import StringIO
from test.testutils import assert_help_output_equals

from simple_parsing import ArgumentParser, field


@dataclass
class InputArgs:
    # Start date from which to collect data about base users. Input in iso format (YYYY-MM-DD).
    # The date is included in the data
    start_date: str = field(alias="s", metadata={"a": "b"})

    # End date for collecting base users. Input in iso format (YYYY-MM-DD). The date is included in the data.
    # Should not be before `start_date`
    end_date: str = field(alias="e")


def test_issue_48():
    parser = ArgumentParser("Prepare input data for training")
    parser.add_arguments(InputArgs, dest="args")
    s = StringIO()
    parser.print_help(file=s)
    s.seek(0)
    output = str(s.read())
    assert_help_output_equals(
        actual=output,
        expected=textwrap.dedent(
            """\
            usage: Prepare input data for training [-h] -s str -e str

            optional arguments:
            -h, --help            show this help message and exit

            InputArgs ['args']:
            InputArgs(start_date:str, end_date:str)

            -s str, --start_date str
                                    Start date from which to collect data about base
                                    users. Input in iso format (YYYY-MM-DD). The date is
                                    included in the data (default: None)
            -e str, --end_date str
                                    End date for collecting base users. Input in iso
                                    format (YYYY-MM-DD). The date is included in the data.
                                    Should not be before `start_date` (default: None)
            """
        ),
    )

    # args = parser.parse_args()



import argparse
from dataclasses import dataclass, field
from typing import Any, Container, List, Tuple, Type, Union

from simple_parsing import ArgumentParser

from .testutils import TestSetup, exits_and_writes_to_stderr, parametrize, raises, xfail


def test_multiple_at_same_dest_throws_error():
    @dataclass
    class SomeClass:
        a: int = 123

    parser = ArgumentParser()
    parser.add_arguments(SomeClass, "some_class")
    with raises(argparse.ArgumentError):
        parser.add_arguments(SomeClass, "some_class")


@xfail(reason="Not sure if this syntax should also be allowed...")
def test_multiple_append_syntax():
    @dataclass
    class SomeClass(TestSetup):
        a: int = 123

    c1, c2, c3 = SomeClass.setup_multiple(3, arguments="--a 1 --a 2 --a 3")
    assert c1.a == 1
    assert c2.a == 2
    assert c3.a == 3


@parametrize("num_instances", [1, 2, 5])
def test_parse_multiple_with_no_arguments_sets_default_value(
    num_instances: int, simple_attribute, silent
):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass(TestSetup):
        a: some_type = expected_value  # type: ignore
        """some docstring for attribute 'a'"""

    classes = SomeClass.setup_multiple(num_instances, "")
    assert len(classes) == num_instances
    for i in range(num_instances):
        c_i = classes[i]
        assert isinstance(c_i, SomeClass)
        assert c_i.a == expected_value
        assert isinstance(c_i.a, some_type)


@parametrize("num_instances", [1, 2, 5])
def test_parse_multiple_with_single_arg_value_sets_that_value_for_all_instances(
    num_instances: int,
    simple_attribute,
    silent,
):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass(TestSetup):
        a: some_type = expected_value  # type: ignore
        """some docstring for attribute 'a'"""

    classes = SomeClass.setup_multiple(num_instances, f"--a {passed_value}")

    assert len(classes) == num_instances
    for i in range(num_instances):
        c_i = classes[i]
        assert isinstance(c_i, SomeClass)
        assert c_i.a == expected_value
        assert isinstance(c_i.a, some_type)


@parametrize(
    "some_type, default_value,  passed_values",
    [
        (int, 123524, [1, 2, 3]),
        (float, 123.456, [4.5, -12.3, 9]),
        (str, "bob", ["random", "triceratops", "cocobongo"]),
        (bool, True, [False, True, False]),
    ],
)
def test_parse_multiple_with_provided_value_for_each_instance(
    some_type: Type, default_value: Any, passed_values: List[Any], silent
):
    @dataclass
    class SomeClass(TestSetup):
        a: some_type = default_value  # type: ignore
        """some docstring for attribute 'a'"""

    # TODO: maybe test out other syntaxes for passing in multiple argument values? (This looks a lot like passing in a list of values..)
    arguments = f"--a {' '.join(str(p) for p in passed_values)}"
    classes = SomeClass.setup_multiple(3, arguments)

    assert len(classes) == 3
    for i in range(3):
        c_i = classes[i]
        assert isinstance(c_i, SomeClass)
        assert c_i.a == passed_values[i]
        assert isinstance(c_i.a, some_type)


@parametrize("some_type", [int, float, str, bool])
def test_parse_multiple_without_required_arguments(some_type: Type):
    @dataclass
    class SomeClass(TestSetup):
        a: some_type  # type: ignore
        """some docstring for attribute 'a'"""

    with exits_and_writes_to_stderr():
        SomeClass.setup_multiple(2, "")


@parametrize("container_type", [List, Tuple])
@parametrize("item_type", [int, float, str, bool])
def test_parse_multiple_without_required_container_arguments(container_type: Type, item_type: Type):
    @dataclass
    class SomeClass(TestSetup):
        a: container_type[item_type]  # type: ignore
        """some docstring for attribute 'a'"""

    with exits_and_writes_to_stderr("the following arguments are required:"):
        _ = SomeClass.setup_multiple(3, "")


@parametrize("container_type", [List, Tuple])
@parametrize("item_type", [int, float, str, bool])
def test_parse_multiple_with_arg_name_without_arg_value(container_type: Type, item_type: Type):
    @dataclass
    class SomeClass(TestSetup):
        a: container_type[item_type]  # type: ignore
        """some docstring for attribute 'a'"""

    with exits_and_writes_to_stderr("expected at least one argument"):
        _ = SomeClass.setup_multiple(3, "--a")


@xfail(reason="BUG: When # of instances to parse is equal to item count in default value.")
@parametrize("num_instances", [1, 2, 3, 5])
@parametrize(
    "container_type, default_value",
    [
        (List[str], ["1", "2", "3"]),
        (List[str], ["1", "2"]),
        (Tuple[str, int], ["1", 2]),
        (List[Union[str, bool]], ["1", False]),
    ],
)
def test_parse_multiple_containers_default_value(
    num_instances: int, container_type: Type[Container], default_value: Container
):
    @dataclass
    class SomeClass(TestSetup):
        a: container_type = field(default_factory=lambda: default_value.copy())  # type: ignore
        """some docstring for attribute 'a'"""

    values = list(SomeClass.setup_multiple(num_instances))
    assert values == [SomeClass(default_value) for i in range(num_instances)]



"""Simple, Elegant Argument parsing.
@author: Fabrice Normandin
"""
from __future__ import annotations

import argparse
import dataclasses
import functools
import itertools
import shlex
import sys
import typing
from argparse import SUPPRESS, Action, HelpFormatter, Namespace, _
from collections import defaultdict
from logging import getLogger
from pathlib import Path
from typing import Any, Callable, Sequence, Type, overload

from simple_parsing.helpers.subgroups import SubgroupKey
from simple_parsing.wrappers.dataclass_wrapper import DataclassWrapperType

from . import utils
from .conflicts import ConflictResolution, ConflictResolver
from .help_formatter import SimpleHelpFormatter
from .helpers.serialization.serializable import read_file
from .utils import (
    Dataclass,
    DataclassT,
    dict_union,
    is_dataclass_instance,
    is_dataclass_type,
)
from .wrappers import DashVariant, DataclassWrapper, FieldWrapper
from .wrappers.field_wrapper import ArgumentGenerationMode, NestedMode

logger = getLogger(__name__)


class ParsingError(RuntimeError, SystemExit):
    pass


class ArgumentParser(argparse.ArgumentParser):
    """Creates an ArgumentParser instance.

    Parameters
    ----------
    - conflict_resolution : ConflictResolution, optional

        What kind of prefixing mechanism to use when reusing dataclasses
        (argument groups).
        For more info, check the docstring of the `ConflictResolution` Enum.

    - add_option_string_dash_variants : DashVariant, optional

        Whether or not to add option_string variants where the underscores in
        attribute names are replaced with dashes.
        For example, when set to DashVariant.UNDERSCORE_AND_DASH,
        "--no-cache" and "--no_cache" can both be used to point to the same
        attribute `no_cache` on some dataclass.

    - argument_generation_mode : ArgumentGenerationMode, optional

        How to generate the arguments. In the ArgumentGenerationMode.FLAT mode,
        the default one, the arguments are flat when possible, ignoring
        their nested structure and including it only on the presence of a
        conflict.

        In the ArgumentGenerationMode.NESTED mode, the arguments are always
        composed reflecting their nested structure.

        In the ArgumentGenerationMode.BOTH mode, both kind of arguments
        are generated.

    - nested_mode : NestedMode, optional

        How to handle argument generation in for nested arguments
        in the modes ArgumentGenerationMode.NESTED and ArgumentGenerationMode.BOTH.
        In the NestedMode.DEFAULT mode, the nested arguments are generated
        reflecting their full 'destination' path from the returning namespace.

        In the NestedMode.WITHOUT_ROOT, the first level is removed. This is useful when
        parser.add_arguments is only called once, and where the same prefix would be shared
        by all arguments. For example, if you have a single dataclass MyArguments and
        you call parser.add_arguments(MyArguments, "args"), the arguments could look like this:
        '--args.input.path --args.output.path'.
        We could prefer to remove the root level in such a case
            so that the arguments get generated as
        '--input.path --output.path'.

    - formatter_class : Type[HelpFormatter], optional

        The formatter class to use. By default, uses
        `simple_parsing.SimpleHelpFormatter`, which is a combination of the
        `argparse.ArgumentDefaultsHelpFormatter`,
        `argparse.MetavarTypeHelpFormatter` and
        `argparse.RawDescriptionHelpFormatter` classes.

    - add_config_path_arg : bool, optional
        When set to `True`, adds a `--config_path` argument, of type Path, which is used to parse

    """

    def __init__(
        self,
        *args,
        parents: Sequence[ArgumentParser] = (),
        add_help: bool = True,
        conflict_resolution: ConflictResolution = ConflictResolution.AUTO,
        add_option_string_dash_variants: DashVariant = DashVariant.AUTO,
        argument_generation_mode=ArgumentGenerationMode.FLAT,
        nested_mode: NestedMode = NestedMode.DEFAULT,
        formatter_class: type[HelpFormatter] = SimpleHelpFormatter,
        add_config_path_arg: bool | None = None,
        config_path: Path | str | Sequence[Path | str] | None = None,
        add_dest_to_option_strings: bool | None = None,
        **kwargs,
    ):
        kwargs["formatter_class"] = formatter_class
        # Pass parents=[] since we override this mechanism below.
        # NOTE: We end up with the same parents.
        super().__init__(*args, parents=[], add_help=False, **kwargs)
        self.conflict_resolution = conflict_resolution

        # constructor arguments for the dataclass instances.
        # (a Dict[dest, [attribute, value]])
        # TODO: Stop using a defaultdict for the very important `self.constructor_arguments`!
        self.constructor_arguments: dict[str, dict[str, Any]] = defaultdict(dict)

        self._conflict_resolver = ConflictResolver(self.conflict_resolution)
        self._wrappers: list[DataclassWrapper] = []

        if add_dest_to_option_strings:
            argument_generation_mode = ArgumentGenerationMode.BOTH

        self._preprocessing_done: bool = False
        self.add_option_string_dash_variants = add_option_string_dash_variants
        self.argument_generation_mode = argument_generation_mode
        self.nested_mode = nested_mode

        FieldWrapper.add_dash_variants = add_option_string_dash_variants
        FieldWrapper.argument_generation_mode = argument_generation_mode
        FieldWrapper.nested_mode = nested_mode
        self._parents = tuple(parents)

        self.add_help = add_help
        if self.add_help:
            prefix_chars = self.prefix_chars
            default_prefix = "-" if "-" in prefix_chars else prefix_chars[0]
            self._help_action = super().add_argument(
                default_prefix + "h",
                default_prefix * 2 + "help",
                action="help",
                default=SUPPRESS,
                help=_("show this help message and exit"),
            )

        self.config_path = Path(config_path) if isinstance(config_path, str) else config_path
        if add_config_path_arg is None:
            # By default, add a config path argument if a config path was passed.
            add_config_path_arg = bool(config_path)
        self.add_config_path_arg = add_config_path_arg

    # TODO: Remove, since the base class already has nicer type hints.
    def add_argument(
        self,
        *name_or_flags: str,
        **kwargs,
    ) -> Action:
        return super().add_argument(
            *name_or_flags,
            **kwargs,
        )

    @overload
    def add_arguments(
        self,
        dataclass: type[DataclassT],
        dest: str,
        *,
        prefix: str = "",
        default: DataclassT | None = None,
        dataclass_wrapper_class: type[DataclassWrapper] = DataclassWrapper,
    ) -> DataclassWrapper[DataclassT]:
        pass

    @overload
    def add_arguments(
        self,
        dataclass: type[Dataclass],
        dest: str,
        *,
        prefix: str = "",
        dataclass_wrapper_class: type[DataclassWrapperType] = DataclassWrapper,
    ) -> DataclassWrapperType:
        pass

    @overload
    def add_arguments(
        self,
        dataclass: DataclassT,
        dest: str,
        *,
        prefix: str = "",
        default: None = None,
        dataclass_wrapper_class: type[DataclassWrapper] = DataclassWrapper,
    ) -> DataclassWrapper[DataclassT]:
        pass

    def add_arguments(
        self,
        dataclass: type[DataclassT] | DataclassT,
        dest: str,
        *,
        prefix: str = "",
        default: DataclassT | None = None,
        dataclass_wrapper_class: type[DataclassWrapperType] = DataclassWrapper,
    ) -> DataclassWrapper[DataclassT] | DataclassWrapperType:
        """Adds command-line arguments for the fields of `dataclass`.

        Parameters
        ----------
        dataclass : Union[Type[Dataclass], Dataclass]
            The dataclass whose fields are to be parsed from the command-line.
            If an instance of a dataclass is given, it is used as the default
            value if none is provided.
        dest : str
            The destination attribute of the `argparse.Namespace` where the
            dataclass instance will be stored after calling `parse_args()`
        prefix : str, optional
            An optional prefix to add prepend to the names of the argparse
            arguments which will be generated for this dataclass.
            This can be useful when registering multiple distinct instances of
            the same dataclass, by default ""
        default : Dataclass, optional
            An instance of the dataclass type to get default values from, by
            default None
        dataclass_wrapper_class : Type[DataclassWrapper], optional
            The type of `DataclassWrapper` to use for this dataclass. This can be used to customize
            how the arguments are generated. However, I'd suggest making a GitHub issue if you find
            yourself using this often.

        Returns
        -------
        The generated DataclassWrapper instance. Feel free to inspect / play around with this if
        you want :)
        """
        if is_dataclass_instance(dataclass):
            if default is not None:
                raise ValueError("Can't use `default` when `dataclass` is a dataclass instance.")
            dataclass = typing.cast(DataclassT, dataclass)
            dataclass_type = type(dataclass)
            default = dataclass
        else:
            if not is_dataclass_type(dataclass):
                raise ValueError(
                    f"`dataclass` should be a dataclass type or instance. Got {dataclass}."
                )
            dataclass = typing.cast(Type[DataclassT], dataclass)
            dataclass_type = dataclass
            default = default

        new_wrapper = self._add_arguments(
            dataclass_type=dataclass_type,
            name=dest,
            prefix=prefix,
            default=default,
            dataclass_wrapper_class=dataclass_wrapper_class,
        )
        self._wrappers.append(new_wrapper)
        return new_wrapper

    def parse_known_args(
        self,
        args: Sequence[str] | None = None,
        namespace: Namespace | None = None,
        attempt_to_reorder: bool = False,
    ):
        # NOTE: since the usual ArgumentParser.parse_args() calls
        # parse_known_args, we therefore just need to overload the
        # parse_known_args method to support both.
        if args is None:
            # args default to the system args
            args = sys.argv[1:]
        else:
            # make sure that args are mutable
            args = list(args)

        # default Namespace built from parser defaults
        if namespace is None:
            namespace = Namespace()
        if self.config_path:
            if isinstance(self.config_path, Path):
                config_paths = [self.config_path]
            else:
                config_paths = self.config_path
            for config_file in config_paths:
                self.set_defaults(config_file)

        if self.add_config_path_arg:
            temp_parser = ArgumentParser(
                add_config_path_arg=False,
                add_help=False,
                add_option_string_dash_variants=FieldWrapper.add_dash_variants,
                argument_generation_mode=FieldWrapper.argument_generation_mode,
                nested_mode=FieldWrapper.nested_mode,
            )
            temp_parser.add_argument(
                "--config_path",
                type=Path,
                nargs="*",
                default=self.config_path,
                help="Path to a config file containing default values to use.",
            )
            args_with_config_path, args = temp_parser.parse_known_args(args)
            config_path = args_with_config_path.config_path

            if config_path is not None:
                config_paths = config_path if isinstance(config_path, list) else [config_path]
                for config_file in config_paths:
                    self.set_defaults(config_file)

            # Adding it here just so it shows up in the help message. The default will be set in
            # the help string.
            self.add_argument(
                "--config_path",
                type=Path,
                default=config_path,
                help="Path to a config file containing default values to use.",
            )

        assert isinstance(args, list)
        self._preprocessing(args=args, namespace=namespace)

        logger.debug(f"Parser {id(self)} is parsing args: {args}, namespace: {namespace}")
        parsed_args, unparsed_args = super().parse_known_args(args, namespace)

        if unparsed_args and self._subparsers and attempt_to_reorder:
            logger.warning(
                f"Unparsed arguments when using subparsers. Will "
                f"attempt to automatically re-order the unparsed arguments "
                f"{unparsed_args}."
            )
            index_in_start = args.index(unparsed_args[0])
            # Simply 'cycle' the args to the right ordering.
            new_start_args = args[index_in_start:] + args[:index_in_start]
            parsed_args, unparsed_args = super().parse_known_args(new_start_args)

        parsed_args = self._postprocessing(parsed_args)
        return parsed_args, unparsed_args

    def add_argument_group(
        self,
        title: str | None = None,
        description: str | None = None,
        prefix_chars=None,
        argument_default=None,
        conflict_handler=None,
    ) -> argparse._ArgumentGroup:
        return super().add_argument_group(
            title=title,
            description=description,
            prefix_chars=prefix_chars or self.prefix_chars,
            argument_default=argument_default or self.argument_default,
            conflict_handler=conflict_handler or self.conflict_handler,
        )

    def print_help(self, file=None, args: Sequence[str] | None = None):
        self._preprocessing(args=list(args) if args else [])
        return super().print_help(file)

    def set_defaults(self, config_path: str | Path | None = None, **kwargs: Any) -> None:
        """Set the default argument values, either from a config file, or from the given kwargs."""
        if config_path:
            defaults = read_file(config_path)
            if self.nested_mode == NestedMode.WITHOUT_ROOT and len(self._wrappers) == 1:
                # The file should have the same format as the command-line args, e.g. contain the
                # fields of the 'root' dataclass directly (e.g. "foo: 123"), rather a dict with
                # "config: foo: 123" where foo is a field of the root dataclass at dest 'config'.
                # Therefore, we add the prefix back here.
                defaults = {self._wrappers[0].dest: defaults}
                # We also assume that the kwargs are passed as foo=123
                kwargs = {self._wrappers[0].dest: kwargs}
            # Also include the values from **kwargs.
            kwargs = dict_union(defaults, kwargs)

        # The kwargs that are set in the dataclasses, rather than on the namespace.
        kwarg_defaults_set_in_dataclasses = {}
        for wrapper in self._wrappers:
            if wrapper.dest in kwargs:
                default_for_dataclass = kwargs[wrapper.dest]

                if isinstance(default_for_dataclass, (str, Path)):
                    default_for_dataclass = read_file(path=default_for_dataclass)
                elif not isinstance(default_for_dataclass, dict) and not dataclasses.is_dataclass(
                    default_for_dataclass
                ):
                    raise ValueError(
                        f"Got a default for field {wrapper.dest} that isn't a dataclass, dict or "
                        f"path: {default_for_dataclass}"
                    )

                # Set the .default attribute on the DataclassWrapper (which also updates the
                # defaults of the fields and any nested dataclass fields).
                wrapper.set_default(default_for_dataclass)

                # It's impossible for multiple wrappers in kwargs to have the same destination.
                assert wrapper.dest not in kwarg_defaults_set_in_dataclasses
                value_for_constructor_arguments = (
                    default_for_dataclass
                    if isinstance(default_for_dataclass, dict)
                    else dataclasses.asdict(default_for_dataclass)
                )
                kwarg_defaults_set_in_dataclasses[wrapper.dest] = value_for_constructor_arguments
                # Remove this from the **kwargs, so they don't get set on the namespace.
                kwargs.pop(wrapper.dest)
        # TODO: Stop using a defaultdict for the very important `self.constructor_arguments`!
        self.constructor_arguments = dict_union(
            self.constructor_arguments,
            kwarg_defaults_set_in_dataclasses,
            dict_factory=lambda: defaultdict(dict),
        )
        # For the rest of the values, use the default argparse behaviour (modifying the
        # self._defaults dictionary).
        super().set_defaults(**kwargs)

    def equivalent_argparse_code(self, args: Sequence[str] | None = None) -> str:
        """Returns the argparse code equivalent to that of `simple_parsing`.

        TODO: Could be fun, pretty sure this is useless though.

        Returns
        -------
        str
            A string containing the auto-generated argparse code.
        """
        self._preprocessing(list(args) if args else [])
        code = "parser = ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)"
        for wrapper in self._wrappers:
            code += "\n"
            code += wrapper.equivalent_argparse_code()
            code += "\n"
        code += "args = parser.parse_args()\n"
        code += "print(args)\n"
        return code

    def _add_arguments(
        self,
        dataclass_type: type[DataclassT],
        name: str,
        *,
        prefix: str = "",
        dataclass_fn: Callable[..., DataclassT] | None = None,
        default: DataclassT | dict | None = None,
        dataclass_wrapper_class: type[DataclassWrapperType] = DataclassWrapper,
        parent: DataclassWrapper | None = None,
    ) -> DataclassWrapper[DataclassT] | DataclassWrapperType:
        assert is_dataclass_type(dataclass_type)
        assert (
            default is None
            or is_dataclass_instance(default)
            or default is argparse.SUPPRESS
            or isinstance(default, dict)
        )
        assert dataclass_fn is None or callable(dataclass_fn)

        for wrapper in self._wrappers:
            if wrapper.dest == name:
                if wrapper.dataclass == dataclass_type:
                    raise argparse.ArgumentError(
                        argument=None,
                        message=f"Destination attribute {name} is already used for "
                        f"dataclass of type {dataclass_type}. Make sure all destinations"
                        f" are unique. (new dataclass type: {dataclass_type})",
                    )
        if not isinstance(dataclass_type, type):
            if default is None:
                default = dataclass_type
            dataclass_type = type(dataclass_type)

        dataclass_fn = dataclass_fn or dataclass_type
        # Create this object that  holds the dataclass we will create arguments for and the
        # arguments that were passed.
        new_wrapper = dataclass_wrapper_class(
            dataclass=dataclass_type,
            name=name,
            prefix=prefix,
            default=default,
            parent=parent,
            dataclass_fn=dataclass_fn,
        )

        if new_wrapper.dest in self._defaults:
            new_wrapper.set_default(self._defaults[new_wrapper.dest])
        if self.nested_mode == NestedMode.WITHOUT_ROOT and all(
            field.name in self._defaults for field in new_wrapper.fields
        ):
            # If we did .set_defaults before we knew what dataclass we're using, then we try to
            # still make use of those defaults:
            new_wrapper.set_default(
                {
                    k: v
                    for k, v in self._defaults.items()
                    if k in [f.name for f in dataclasses.fields(new_wrapper.dataclass)]
                }
            )

        return new_wrapper

    def _preprocessing(self, args: Sequence[str] = (), namespace: Namespace | None = None) -> None:
        """Resolve potential conflicts, resolve subgroups, and add all the arguments."""
        logger.debug("\nPREPROCESSING\n")

        if self._preprocessing_done:
            return

        args = list(args)

        wrapped_dataclasses = self._wrappers.copy()
        # Fix the potential conflicts between dataclass fields with the same names.
        wrapped_dataclasses = self._conflict_resolver.resolve_and_flatten(wrapped_dataclasses)

        wrapped_dataclasses, chosen_subgroups = self._resolve_subgroups(
            wrappers=wrapped_dataclasses, args=args, namespace=namespace
        )

        # NOTE: We keep the subgroup fields in their dataclasses so they show up with the other
        # arguments.
        wrapped_dataclasses = _flatten_wrappers(wrapped_dataclasses)

        # Create one argument group per dataclass
        for wrapped_dataclass in wrapped_dataclasses:
            logger.debug(
                f"Parser {id(self)} is Adding arguments for dataclass: {wrapped_dataclass.dataclass} "
                f"at destinations {wrapped_dataclass.destinations}"
            )
            wrapped_dataclass.add_arguments(parser=self)

        self._wrappers = wrapped_dataclasses
        # Save this so we don't re-add all the arguments.
        self._preprocessing_done = True

    def _postprocessing(self, parsed_args: Namespace) -> Namespace:
        """Process the namespace by extract the fields and creating the objects.

        Instantiate the dataclasses from the parsed arguments and set them at
        their destination attribute in the namespace.

        Parameters
        ----------
        parsed_args : Namespace
            the result of calling `super().parse_args(...)` or
            `super().parse_known_args(...)`.
            TODO: Try and maybe return a nicer, typed version of parsed_args.


        Returns
        -------
        Namespace
            The original Namespace, with all the arguments corresponding to the
            dataclass fields removed, and with the added dataclass instances.
            Also keeps whatever arguments were added in the traditional fashion,
            i.e. with `parser.add_argument(...)`.
        """
        logger.debug("\nPOST PROCESSING\n")
        logger.debug(f"(raw) parsed args: {parsed_args}")

        self._remove_subgroups_from_namespace(parsed_args)
        # create the constructor arguments for each instance by consuming all
        # the relevant attributes from `parsed_args`
        wrappers = _flatten_wrappers(self._wrappers)

        constructor_arguments = self.constructor_arguments.copy()
        for wrapper in wrappers:
            for destination in wrapper.destinations:
                constructor_arguments.setdefault(destination, {})

        parsed_args, constructor_arguments = self._fill_constructor_arguments_with_fields(
            parsed_args, wrappers=wrappers, initial_constructor_arguments=constructor_arguments
        )
        parsed_args = self._instantiate_dataclasses(
            parsed_args, wrappers=wrappers, constructor_arguments=constructor_arguments
        )
        return parsed_args

    def _resolve_subgroups(
        self,
        wrappers: list[DataclassWrapper],
        args: list[str],
        namespace: Namespace | None = None,
    ) -> tuple[list[DataclassWrapper], dict[str, str]]:
        """Iteratively add and resolve all the choice of argument subgroups, if any.

        This modifies the wrappers in-place, by possibly adding children to the wrappers in the
        list.
        Returns a list with the modified wrappers.

        Each round does the following:
        1.  Resolve any conflicts using the conflict resolver. Two subgroups at the same nesting
            level, with the same name, get a different prefix, for example "--generator.optimizer"
            and "--discriminator.optimizer".
        2.  Add all the subgroup choice arguments to a parser.
        3.  Add the chosen dataclasses to the list of dataclasses to parse later in the main
            parser. This is done by adding wrapping the dataclass and adding it to the `wrappers`
            list.
        """

        unresolved_subgroups = _get_subgroup_fields(wrappers)
        # Dictionary of the subgroup choices that were resolved (key: subgroup dest, value: chosen
        # subgroup name).
        resolved_subgroups: dict[str, SubgroupKey] = {}

        if not unresolved_subgroups:
            # No subgroups to parse.
            return wrappers, {}

        # Use a temporary parser, to avoid parsing "vanilla argparse" arguments of `self` multiple
        # times.
        subgroup_choice_parser = argparse.ArgumentParser(
            add_help=False,
            # conflict_resolution=self.conflict_resolution,
            # add_option_string_dash_variants=self.add_option_string_dash_variants,
            # argument_generation_mode=self.argument_generation_mode,
            # nested_mode=self.nested_mode,
            formatter_class=self.formatter_class,
            # add_config_path_arg=self.add_config_path_arg,
            # config_path=self.config_path,
            # NOTE: We disallow abbreviations for subgroups for now. This prevents potential issues
            # for example if you have —a_or_b and A has a field —a then it will error out if you
            # pass —a=1 because 1 isn’t a choice for the a_or_b argument (because --a matches it
            # with the abbreviation feature turned on).
            allow_abbrev=False,
        )

        for current_nesting_level in itertools.count():
            # Do rounds of parsing with just the subgroup arguments, until all the subgroups
            # are resolved to a dataclass type.
            logger.debug(
                f"Starting subgroup parsing round {current_nesting_level}: {list(unresolved_subgroups.keys())}"
            )
            # Add all the unresolved subgroups arguments.
            for dest, subgroup_field in unresolved_subgroups.items():
                flags = subgroup_field.option_strings
                argument_options = subgroup_field.arg_options

                if subgroup_field.subgroup_default is dataclasses.MISSING:
                    assert argument_options["required"]
                else:
                    assert argument_options["default"] is subgroup_field.subgroup_default
                    assert not is_dataclass_instance(argument_options["default"])

                # TODO: Do we really need to care about this "SUPPRESS" stuff here?
                if argparse.SUPPRESS in subgroup_field.parent.defaults:
                    assert argument_options["default"] is argparse.SUPPRESS
                    argument_options["default"] = argparse.SUPPRESS

                logger.debug(
                    f"Adding subgroup argument: add_argument(*{flags} **{str(argument_options)})"
                )
                subgroup_choice_parser.add_argument(*flags, **argument_options)

            # Parse `args` repeatedly until all the subgroup choices are resolved.
            parsed_args, unused_args = subgroup_choice_parser.parse_known_args(
                args=args, namespace=namespace
            )
            logger.debug(
                f"Nesting level {current_nesting_level}: args: {args}, "
                f"parsed_args: {parsed_args}, unused_args: {unused_args}"
            )

            for dest, subgroup_field in list(unresolved_subgroups.items()):
                # NOTE: There should always be a parsed value for the subgroup argument on the
                # namespace. This is because we added all the subgroup arguments before we get
                # here.
                subgroup_dict = subgroup_field.subgroup_choices
                chosen_subgroup_key: SubgroupKey = getattr(parsed_args, dest)
                assert chosen_subgroup_key in subgroup_dict

                # Changing the default value of the (now parsed) field for the subgroup choice,
                # just so it shows (default: {chosen_subgroup_key}) on the command-line.
                # Note: This really isn't required, we could have it just be the default value, but
                # it seems a bit more consistent with us then showing the --help string for the
                # chosen dataclass type (as we're doing below).
                # subgroup_field.set_default(chosen_subgroup_key)
                logger.debug(
                    f"resolved the subgroup at {dest!r}: will use the subgroup at key "
                    f"{chosen_subgroup_key!r}"
                )

                default_or_dataclass_fn = subgroup_dict[chosen_subgroup_key]
                if is_dataclass_instance(default_or_dataclass_fn):
                    # The chosen value in the subgroup dict is a frozen dataclass instance.
                    default = default_or_dataclass_fn
                    dataclass_fn = functools.partial(dataclasses.replace, default)
                    dataclass_type = type(default)
                else:
                    default = None
                    dataclass_fn = default_or_dataclass_fn
                    dataclass_type = subgroup_field.field.metadata["subgroup_dataclass_types"][
                        chosen_subgroup_key
                    ]

                assert default is None or is_dataclass_instance(default)
                assert callable(dataclass_fn)
                assert is_dataclass_type(dataclass_type)

                name = dest.split(".")[-1]
                parent_dataclass_wrapper = subgroup_field.parent
                # NOTE: Using self._add_arguments so it returns the modified wrapper and doesn't
                # affect the `self._wrappers` list.
                new_wrapper = self._add_arguments(
                    dataclass_type=dataclass_type,
                    name=name,
                    dataclass_fn=dataclass_fn,
                    default=default,
                    parent=parent_dataclass_wrapper,
                )
                # Make the new wrapper a child of the class which contains the field.
                # - it isn't already a child
                # - it's parent is the parent dataclass wrapper
                # - the parent is already in the tree of DataclassWrappers.
                assert new_wrapper not in parent_dataclass_wrapper._children
                parent_dataclass_wrapper._children.append(new_wrapper)
                assert new_wrapper.parent is parent_dataclass_wrapper
                assert parent_dataclass_wrapper in _flatten_wrappers(wrappers)
                assert new_wrapper in _flatten_wrappers(wrappers)

                # Mark this subgroup as resolved.
                unresolved_subgroups.pop(dest)
                resolved_subgroups[dest] = chosen_subgroup_key
                # TODO: Should we remove the FieldWrapper for the subgroups now that it's been
                # resolved?

            # Find the new subgroup fields that weren't resolved before.
            # TODO: What if a name conflict occurs between a subgroup field and one of the new
            # fields below it? For example, something like --model model_a (and inside the `ModelA`
            # dataclass, there's a field called `model`. Then, this will cause a conflict!)
            # For now, I'm just going to wait and see how this plays out. I'm hoping that the
            # auto conflict resolution shouldn't run into any issues in this case.

            wrappers = self._conflict_resolver.resolve(wrappers)

            all_subgroup_fields = _get_subgroup_fields(wrappers)
            unresolved_subgroups = {
                k: v for k, v in all_subgroup_fields.items() if k not in resolved_subgroups
            }
            logger.debug(f"All subgroups: {list(all_subgroup_fields.keys())}")
            logger.debug(f"Resolved subgroups: {resolved_subgroups}")
            logger.debug(f"Unresolved subgroups: {list(unresolved_subgroups.keys())}")

            if not unresolved_subgroups:
                logger.debug("Done parsing all the subgroups!")
                break
            else:
                logger.debug(
                    f"Done parsing a round of subparsers at nesting level "
                    f"{current_nesting_level}. Moving to the next round which has "
                    f"{len(unresolved_subgroups)} unresolved subgroup choices."
                )
        return wrappers, resolved_subgroups

    def _remove_subgroups_from_namespace(self, parsed_args: argparse.Namespace) -> None:
        """Removes the subgroup choice results from the namespace.
        Modifies the namespace in-place.
        """
        # find all subgroup fields
        subgroup_fields = _get_subgroup_fields(self._wrappers)

        if not subgroup_fields:
            return
        # IDEA: Store the choices in a `subgroups` dict on the namespace.
        if not hasattr(parsed_args, "subgroups"):
            parsed_args.subgroups = {}

        for dest in subgroup_fields:
            chosen_value = getattr(parsed_args, dest)
            parsed_args.subgroups[dest] = chosen_value
            delattr(parsed_args, dest)

    def _instantiate_dataclasses(
        self,
        parsed_args: argparse.Namespace,
        wrappers: list[DataclassWrapper],
        constructor_arguments: dict[str, dict[str, Any]],
    ) -> argparse.Namespace:
        """Create the instances set them at their destination in the namespace.

        We now have all the constructor arguments for each instance.
        We can now sort out the dependencies, create the instances, and set them
        as attributes of the Namespace.

        Since the dataclasses might have nested children, and we need to pass
        all the constructor arguments when calling the dataclass constructors,
        we create the instances in a "bottom-up" fashion, creating the deepest
        objects first, and then setting their value in the
        `constructor_arguments` dict.

        Parameters
        ----------
        parsed_args : argparse.Namespace
            The 'raw' Namespace that is produced by `parse_args`.

        wrappers : list[DataclassWrapper]
            The (assumed flattened) list of dataclass wrappers that were created with
            `add_arguments`.

        constructor_arguments : dict[str, dict[str, Any]]
            The partially populated dict of constructor arguments for each dataclass. This will be
            consumed in order to create the dataclass instances for each DataclassWrapper.

        Returns
        -------
        argparse.Namespace
            The transformed namespace with the instances set at their
            corresponding destinations.
        """
        constructor_arguments = constructor_arguments.copy()
        # FIXME: There's a bug here happening with the `ALWAYS_MERGE` case: The namespace has the
        # values, but the constructor arguments dict doesn't.

        if self.conflict_resolution != ConflictResolution.ALWAYS_MERGE:
            assert len(wrappers) == len(constructor_arguments), "should have one dict per wrapper"

        # sort the wrappers so as to construct the leaf nodes first.
        sorted_dc_wrappers: list[DataclassWrapper] = sorted(
            wrappers, key=lambda w: w.nesting_level, reverse=True
        )
        assert len(sorted_dc_wrappers) == len(set(sorted_dc_wrappers))

        for dc_wrapper in sorted_dc_wrappers:
            logger.info(f"Instantiating the wrapper with destinations {dc_wrapper.destinations}")

            for destination in dc_wrapper.destinations:
                logger.info(f"Instantiating the dataclass at destination {destination}")
                # Instantiate the dataclass by passing the constructor arguments
                # to the constructor.
                constructor = dc_wrapper.dataclass_fn
                constructor_args = constructor_arguments.pop(destination)
                # If the dataclass wrapper is marked as 'optional' and all the
                # constructor args are None, then the instance is None.
                value_for_dataclass_field: Any | dict[str, Any] | None
                if argparse.SUPPRESS in dc_wrapper.defaults:
                    if constructor_args == {}:
                        value_for_dataclass_field = None
                    else:
                        # Don't create the dataclass instance. Instead, keep the value as a dict.
                        value_for_dataclass_field = constructor_args
                else:
                    value_for_dataclass_field = _create_dataclass_instance(
                        dc_wrapper, constructor, constructor_args
                    )

                if argparse.SUPPRESS in dc_wrapper.defaults and value_for_dataclass_field is None:
                    logger.debug(
                        f"Suppressing entire destination {destination} because none of its"
                        f"subattributes were specified on the command line."
                    )

                elif dc_wrapper.parent is not None:
                    parent_key, attr = utils.split_dest(destination)
                    logger.debug(
                        f"Setting a value of {value_for_dataclass_field} at attribute {attr} in "
                        f"parent at key {parent_key}."
                    )
                    constructor_arguments[parent_key][attr] = value_for_dataclass_field

                elif not hasattr(parsed_args, destination):
                    logger.debug(
                        f"setting attribute '{destination}' on the Namespace "
                        f"to a value of {value_for_dataclass_field}"
                    )
                    setattr(parsed_args, destination, value_for_dataclass_field)

                else:
                    # There is a collision: namespace already has an entry at this destination.
                    existing = getattr(parsed_args, destination)
                    if dc_wrapper.dest in self._defaults:
                        logger.debug(
                            f"Overwriting defaults in the namespace at destination '{destination}' "
                            f"on the Namespace ({existing}) to a value of {value_for_dataclass_field}"
                        )
                        setattr(parsed_args, destination, value_for_dataclass_field)
                    else:
                        raise RuntimeError(
                            f"Namespace should not already have a '{destination}' "
                            f"attribute!\n"
                            f"The value would be overwritten:\n"
                            f"- existing value: {existing}\n"
                            f"- new value:      {value_for_dataclass_field}"
                        )

        # We should be consuming all the constructor arguments.
        assert not constructor_arguments

        return parsed_args

    def _fill_constructor_arguments_with_fields(
        self,
        parsed_args: argparse.Namespace,
        wrappers: list[DataclassWrapper],
        initial_constructor_arguments: dict[str, dict[str, Any]],
    ) -> tuple[argparse.Namespace, dict[str, dict[str, Any]]]:
        """Create the constructor arguments for each instance.

        Creates the arguments by consuming all the attributes from
        `parsed_args`.
        Here we imitate a custom action, by having the FieldWrappers be
        callables that set their value in the `constructor_args` attribute.

        Parameters
        ----------
        parsed_args : argparse.Namespace
            the argparse.Namespace returned from super().parse_args().

        wrappers : list[DataclassWrapper]
            The (assumed flattened) list of dataclass wrappers that were created with
            `add_arguments`.

        constructor_arguments : dict[str, dict[str, Any]]
            The dict of constructor arguments to create for each dataclass. This will be filled by
            each FieldWrapper.

        Returns
        -------
        argparse.Namespace
            The namespace, without the consumed arguments.
        """

        if self.conflict_resolution != ConflictResolution.ALWAYS_MERGE:
            assert len(wrappers) == len(
                initial_constructor_arguments
            ), "should have one dict per wrapper"

        # The output
        constructor_arguments = initial_constructor_arguments.copy()

        parsed_arg_values = vars(parsed_args)
        deleted_values: dict[str, Any] = {}

        for wrapper in wrappers:
            for field in wrapper.fields:
                if argparse.SUPPRESS in wrapper.defaults and field.dest not in parsed_args:
                    continue

                if field.is_subgroup:
                    # Skip the subgroup fields, since we added a child DataclassWrapper for them.
                    logger.debug(f"Not calling the subgroup FieldWrapper for dest {field.dest}")
                    continue

                if not field.field.init:
                    # The field isn't an argument of the dataclass constructor.
                    continue

                # NOTE: If the field is reused (when using the ConflictResolution.ALWAYS_MERGE
                # strategy), then we store the multiple values in the `dest` of the first field.
                # They are they distributed in `constructor_arguments` using the
                # `field.destinations`, which gives the destination for each value.
                values = parsed_arg_values.pop(field.dest, field.default)
                deleted_values[field.dest] = values

                # call the "action" for the given attribute. This sets the right
                # value in the `constructor_arguments` dictionary.
                field(
                    parser=self,
                    namespace=parsed_args,
                    values=values,
                    constructor_arguments=constructor_arguments,
                )

        # "Clean up" the Namespace by returning a new Namespace without the
        # consumed attributes.
        leftover_args = argparse.Namespace(**parsed_arg_values)
        if deleted_values:
            logger.debug(f"deleted values: {deleted_values}")
            logger.debug(f"leftover args: {leftover_args}")

        return leftover_args, constructor_arguments


# TODO: Change the order of arguments to put `args` as the second argument.
def parse(
    config_class: type[DataclassT],
    config_path: Path | str | None = None,
    args: str | Sequence[str] | None = None,
    default: DataclassT | None = None,
    dest: str = "config",
    *,
    prefix: str = "",
    nested_mode: NestedMode = NestedMode.WITHOUT_ROOT,
    conflict_resolution: ConflictResolution = ConflictResolution.AUTO,
    add_option_string_dash_variants: DashVariant = DashVariant.AUTO,
    argument_generation_mode=ArgumentGenerationMode.FLAT,
    formatter_class: type[HelpFormatter] = SimpleHelpFormatter,
    add_config_path_arg: bool | None = None,
    **kwargs,
) -> DataclassT:
    """Parse the given dataclass from the command-line.

    See the `ArgumentParser` constructor for more details on the arguments (they are the same here
    except for `nested_mode`, which has a different default value).

    If `config_path` is passed, loads the values from that file and uses them as defaults.
    """
    parser = ArgumentParser(
        nested_mode=nested_mode,
        add_help=True,
        # add_config_path_arg=None,
        config_path=config_path,
        conflict_resolution=conflict_resolution,
        add_option_string_dash_variants=add_option_string_dash_variants,
        argument_generation_mode=argument_generation_mode,
        formatter_class=formatter_class,
        add_config_path_arg=add_config_path_arg,
        **kwargs,
    )

    parser.add_arguments(config_class, prefix=prefix, dest=dest, default=default)

    if isinstance(args, str):
        args = shlex.split(args)
    parsed_args = parser.parse_args(args)

    config: Dataclass = getattr(parsed_args, dest)
    return config


def parse_known_args(
    config_class: type[Dataclass],
    config_path: Path | str | None = None,
    args: str | Sequence[str] | None = None,
    default: Dataclass | None = None,
    dest: str = "config",
    attempt_to_reorder: bool = False,
    *,
    nested_mode: NestedMode = NestedMode.WITHOUT_ROOT,
    conflict_resolution: ConflictResolution = ConflictResolution.AUTO,
    add_option_string_dash_variants: DashVariant = DashVariant.AUTO,
    argument_generation_mode=ArgumentGenerationMode.FLAT,
    formatter_class: type[HelpFormatter] = SimpleHelpFormatter,
    add_config_path_arg: bool | None = None,
) -> tuple[Dataclass, list[str]]:
    """Parse the given dataclass from the command-line, returning the leftover arguments.

    See the `ArgumentParser` constructor for more details on the arguments (they are the same here
    except for `nested_mode`, which has a different default value).

    If `config_path` is passed, loads the values from that file and uses them as defaults.
    """

    if isinstance(args, str):
        args = shlex.split(args)
    parser = ArgumentParser(
        nested_mode=nested_mode,
        add_help=True,
        # add_config_path_arg=None,
        config_path=config_path,
        conflict_resolution=conflict_resolution,
        add_option_string_dash_variants=add_option_string_dash_variants,
        argument_generation_mode=argument_generation_mode,
        formatter_class=formatter_class,
        add_config_path_arg=add_config_path_arg,
    )
    parser.add_arguments(config_class, dest=dest, default=default)
    parsed_args, unknown_args = parser.parse_known_args(args, attempt_to_reorder=attempt_to_reorder)
    config: Dataclass = getattr(parsed_args, dest)
    return config, unknown_args


def _get_subgroup_fields(wrappers: list[DataclassWrapper]) -> dict[str, FieldWrapper]:
    subgroup_fields = {}
    all_wrappers = _flatten_wrappers(wrappers)
    for wrapper in all_wrappers:
        for field in wrapper.fields:
            if field.is_subgroup:
                assert field not in subgroup_fields.values()
                subgroup_fields[field.dest] = field
    return subgroup_fields


def _remove_duplicates(wrappers: list[DataclassWrapper]) -> list[DataclassWrapper]:
    return list(set(wrappers))


def _assert_no_duplicates(wrappers: list[DataclassWrapper]) -> None:
    if len(wrappers) != len(set(wrappers)):
        raise RuntimeError(
            "Duplicate wrappers found! This is a potentially nasty bug on our "
            "part. Please make an issue at https://www.github.com/lebrice/SimpleParsing/issues "
        )


def _flatten_wrappers(wrappers: list[DataclassWrapper]) -> list[DataclassWrapper]:
    """Takes a list of nodes, returns a flattened list of all nodes in the tree."""
    _assert_no_duplicates(wrappers)
    roots_only = _unflatten_wrappers(wrappers)
    return sum(([w] + list(w.descendants) for w in roots_only), [])


def _unflatten_wrappers(wrappers: list[DataclassWrapper]) -> list[DataclassWrapper]:
    """Given a list of nodes in one or more trees, returns only the root nodes.

    In our context, this is all the dataclass arg groups that were added with
    `parser.add_arguments`.
    """
    _assert_no_duplicates(wrappers)
    return [w for w in wrappers if w.parent is None]


def _create_dataclass_instance(
    wrapper: DataclassWrapper[DataclassT],
    constructor: Callable[..., DataclassT],
    constructor_args: dict[str, Any],
) -> DataclassT | None:

    # Check if the dataclass annotation is marked as Optional.
    # In this case, if no arguments were passed, and the default value is None, then return
    # None.
    # TODO: (BUG!) This doesn't distinguish the case where the defaults are passed via the
    # command-line from the case where no arguments are passed at all!
    if wrapper.optional and wrapper.default is None:
        for field_wrapper in wrapper.fields:

            arg_value = constructor_args[field_wrapper.name]
            default_value = field_wrapper.default
            logger.debug(
                f"field {field_wrapper.name}, arg value: {arg_value}, "
                f"default value: {default_value}"
            )
            if arg_value != default_value:
                # Value is not the default value, so an argument must have been passed.
                # Break, and return the instance.
                break
        else:
            logger.debug(f"All fields for {wrapper.dest} were either at their default, or None.")
            return None
    logger.debug(f"Calling constructor: {constructor}(**{constructor_args})")
    return constructor(**constructor_args)



from dataclasses import dataclass
from typing import List

from simple_parsing import field

from .testutils import TestSetup


def test_single_posarg():
    @dataclass
    class Foo(TestSetup):
        output_dir: str = field(positional=True)
        extra_flag: bool = False

    foo = Foo.setup("/bob --extra_flag")
    assert foo.output_dir == "/bob"
    assert foo.extra_flag


def test_repeated_posarg():
    @dataclass
    class Foo(TestSetup):
        output_dir: List[str] = field(positional=True)
        extra_flag: bool = False

    # Here we see why 'invoke' wrote their own parser. Doesn't seem obvious how to explain to argparse that
    # --extra_flag /cherry (which is a little ambiguous) or --extra_flag True /cherry is something we'd like to allow.
    foo = Foo.setup("/alice /bob /cherry --extra_flag")
    assert foo.output_dir == ["/alice", "/bob", "/cherry"]
    assert foo.extra_flag



import math
import random
from abc import abstractmethod
from dataclasses import dataclass
from typing import (
    Any,
    Generic,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)

numpy_installed = False
try:
    import numpy as np

    numpy_installed = True
except ImportError:
    pass


T = TypeVar("T")


@dataclass  # type: ignore
class Prior(Generic[T]):
    def __post_init__(self):
        if numpy_installed:
            self.np_rng = np.random
        else:
            self.rng: random.Random = random.Random()

    @abstractmethod
    def sample(self) -> T:
        pass

    def seed(self, seed: Optional[int]) -> None:
        # Should this seed this individual prior?
        if numpy_installed:
            self.np_rng = np.random.RandomState(seed)
        else:
            self.rng = random.Random(seed)

    @abstractmethod
    def get_orion_space_string(self) -> str:
        """Gets the 'Orion-formatted space string' for this Prior object."""

    @abstractmethod
    def __contains__(self, v: Union[T, Any]) -> bool:
        pass


@dataclass
class NormalPrior(Prior):
    mu: float = 0.0
    sigma: float = 1.0
    discrete: bool = False
    default: Optional[float] = None
    shape: Union[int, Tuple[int, ...]] = None

    def __post_init__(self):
        super().__post_init__()
        if self.shape:
            if isinstance(self.default, (int, float)):
                self.default = [self.default for _ in range(self.shape)]

    def sample(self) -> Union[float, int]:
        if self.shape:
            assert isinstance(self.shape, int), "only support int shape for now."
            if numpy_installed:
                return self.np_rng.normal(self.mu, self.sigma, size=self.shape)
            elif isinstance(self.shape, int):
                _shape = self.shape
                self.shape = None
                values = [self.sample() for _ in range(_shape)]
                self.shape = _shape
                return values
            else:
                raise NotImplementedError(self.shape)
        if numpy_installed:
            value = self.np_rng.normal(self.mu, self.sigma)
        else:
            value = self.rng.normalvariate(self.mu, self.sigma)
        if self.discrete:
            return round(value)
        return value

    def get_orion_space_string(self) -> str:
        raise NotImplementedError(
            "TODO: Add this for the normal prior, didn't check how its done in " "Orion yet."
        )

    def __contains__(self, v: Union[T, Any]) -> bool:
        # TODO: For normal priors, I guess we only really check if the value is a float?
        return isinstance(v, (int, float))


@dataclass
class UniformPrior(Prior):
    min: float = 0.0
    max: float = 1.0
    discrete: bool = False
    default: Optional[float] = None
    shape: Union[int, Tuple[int, ...]] = None

    def __post_init__(self):
        super().__post_init__()
        assert self.min <= self.max
        if self.shape:
            if isinstance(self.default, (int, float)):
                self.default = [self.default for _ in range(self.shape)]

    def sample(self) -> Union[float, int]:
        # TODO: add support for enums?
        if self.shape:
            assert isinstance(self.shape, int), "only support int shape for now."
            if numpy_installed:
                values = self.np_rng.uniform(self.min, self.max, size=self.shape)
                if self.discrete:
                    values = np.round(values)
                    values = values.astype(int)
                return values
            elif isinstance(self.shape, int):
                _shape = self.shape
                self.shape = None
                values = [self.sample() for _ in range(_shape)]
                self.shape = _shape
                return values
            else:
                raise NotImplementedError(self.shape)
        if numpy_installed:
            value = self.np_rng.uniform(self.min, self.max)
        else:
            value = self.rng.uniform(self.min, self.max)
        if self.discrete:
            return round(value)
        return value

    def get_orion_space_string(self) -> str:
        string = f"uniform({self.min}, {self.max}"
        if self.discrete:
            string += ", discrete=True"
        if self.default is not None:
            string += f", default_value={self.default}"
        if self.shape is not None:
            string += f", shape={self.shape}"
        string += ")"
        return string

    def __contains__(self, v: Union[T, Any]) -> bool:
        # TODO: Include the max value here? or not?
        return isinstance(v, (int, float)) and (self.min <= v < self.max)


@dataclass
class CategoricalPrior(Prior[T]):
    choices: List[T]
    probabilities: Optional[List[float]] = None
    default_value: Optional[T] = None

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.choices, dict):
            choices = []
            self.probabilities = []
            for k, v in self.choices.items():
                choices.append(k)
                assert isinstance(v, (int, float)), "probs should be int or float"
                self.probabilities.append(v)

    @overload
    def sample(self, n: int) -> List[T]:
        ...

    @overload
    def sample(self) -> T:
        ...

    def sample(self, n: int = None) -> Union[T, List[T]]:
        assert self.choices
        # n = n or 1
        # assert isinstance(n, int), n
        choices: List = []
        probabilities: List[float] = []
        if isinstance(self.choices, dict):
            for k, v in self.choices.items():
                choices.append(k)
                probabilities.append(v)
        else:
            choices = self.choices
            probabilities = self.probabilities

        print(choices, n, probabilities)
        if numpy_installed:
            s = self.np_rng.choice(choices, size=n, p=probabilities)
            samples = [(s_i.item() if isinstance(s_i, np.ndarray) else s_i) for s_i in s]
        else:
            samples = self.rng.choices(choices, weights=probabilities, k=n or 1)

        return samples[0] if n in {None, 1} else samples

    def get_orion_space_string(self) -> str:
        string = "choices("
        if self.probabilities:
            prob_dict = dict(zip(self.choices, self.probabilities))
            assert sum(self.probabilities) == 1, "probs should sum to 1."
            # BUG: Seems like orion still samples entries, even if they have zero
            # probability!
            # TODO: Remove the entries that have zero prob?
            prob_dict = {k: v for k, v in prob_dict.items() if v > 0}
            string += str(prob_dict)
        else:
            string += str(self.choices)
        if self.default_value is not None:
            assert isinstance(self.default_value, (int, str, float))
            default_value_str = str(self.default_value)
            if isinstance(self.default_value, str):
                default_value_str = f"'{self.default_value}'"
            string += f", default_value={default_value_str}"
        string += ")"
        return string

    def __contains__(self, v: Union[T, Any]) -> bool:
        return v in self.choices


@dataclass
class LogUniformPrior(Prior):
    min: float = 1e-3
    max: float = 1e3
    base: float = math.e
    discrete: bool = False
    default: Optional[float] = None
    shape: Union[int, Tuple[int, ...]] = None

    def __post_init__(self):
        super().__post_init__()
        if self.shape:
            if isinstance(self.default, (int, float)):
                self.default = [self.default for _ in range(self.shape)]

    def sample(self) -> float:
        # TODO: Might not be 100% numerically stable.
        assert self.min > 0, "min of LogUniform can't be negative!"
        assert self.min < self.max, "max should be greater than min!"
        if self.shape:
            assert isinstance(self.shape, int), "only support in shape for now."
            if numpy_installed:
                log_vals = self.np_rng.uniform(self.log_min, self.log_max, size=self.shape)
                values = np.power(self.base, log_vals)
                if self.discrete:
                    values = np.round(values)
                return values
            elif isinstance(self.shape, int):
                _shape = self.shape
                self.shape = None
                values = [self.sample() for _ in range(_shape)]
                self.shape = _shape
                return values
            else:
                raise NotImplementedError(self.shape)
        if numpy_installed:
            log_val = self.np_rng.uniform(self.log_min, self.log_max)
        else:
            log_val = self.rng.uniform(self.log_min, self.log_max)
        value = math.pow(self.base, log_val)
        if self.discrete:
            return round(value)
        return value

    @property
    def log_min(self) -> Union[int, float]:
        if numpy_installed:
            if self.base in {np.e, math.e}:
                log_min = np.log(self.min)
            else:
                log_min = np.log(self.min)
        else:
            if self.base is math.e:
                log_min = math.log(self.min)
            else:
                log_min = math.log(self.min, self.base)
        assert isinstance(log_min, (int, float))
        return log_min

    @property
    def log_max(self) -> Union[int, float]:
        if numpy_installed:
            if self.base in {math.e, np.e}:
                log_max = np.log(self.max)
            else:
                log_max = np.log(self.max) / np.log(self.base)
        else:
            if self.base is math.e:
                log_max = math.log(self.max)
            else:
                log_max = math.log(self.max, self.base)
        assert isinstance(log_max, (int, float))
        return log_max

    def get_orion_space_string(self) -> str:
        def format_power(value: float, log_value: float):
            if isinstance(value, int) or value.is_integer():
                return int(value)
            elif isinstance(log_value, int) or log_value.is_integer():
                log_value = int(log_value)
                if self.base == np.e:
                    return f"np.exp({int(log_value)})"
                elif self.base == 10:
                    return f"{value:.2e}"
            if math.log10(value).is_integer():
                return f"{value:.0e}"
            else:
                return f"{value:g}"

        min_str = format_power(self.min, self.log_min)
        max_str = format_power(self.max, self.log_max)
        string = f"loguniform({min_str}, {max_str}"
        if self.discrete:
            string += ", discrete=True"
        if self.default is not None:
            string += f", default_value={self.default}"
        if self.shape is not None:
            string += f", shape={self.shape}"
        string += ")"
        return string

    def __contains__(self, v: Union[T, Any]) -> bool:
        if self.shape:
            assert isinstance(self.shape, int), "only support int shape for now."
            mins: Sequence[float]
            if isinstance(self.min, (int, float)):
                mins = [self.min for _ in range(self.shape)]
            else:
                mins = self.min

            maxes: Sequence[float]
            if isinstance(self.max, (int, float)):
                maxes = [self.max for _ in range(self.shape)]
            else:
                maxes = self.max

            return all(
                isinstance(v_i, (int, float)) and mins[i] <= v_i < maxes[i]
                for i, v_i in enumerate(v)
            )
        return isinstance(v, (int, float)) and (self.min <= v < self.max)



from __future__ import annotations

import functools
import inspect
import typing
from dataclasses import _MISSING_TYPE, MISSING
from enum import Enum
from logging import getLogger as get_logger
from typing import Any, Callable, TypeVar, Union, overload

from typing_extensions import TypeAlias

from simple_parsing.utils import DataclassT, is_dataclass_instance, is_dataclass_type

logger = get_logger(__name__)

SubgroupKey: TypeAlias = Union[str, int, bool, Enum]

Key = TypeVar("Key", str, int, bool, Enum)


@overload
def subgroups(
    subgroups: dict[Key, DataclassT | type[DataclassT] | functools.partial[DataclassT]],
    *args,
    default: Key | DataclassT,
    default_factory: _MISSING_TYPE = MISSING,
    **kwargs,
) -> DataclassT:
    ...


@overload
def subgroups(
    subgroups: dict[Key, DataclassT | type[DataclassT] | functools.partial[DataclassT]],
    *args,
    default: _MISSING_TYPE = MISSING,
    default_factory: type[DataclassT] | functools.partial[DataclassT],
    **kwargs,
) -> DataclassT:
    ...


@overload
def subgroups(
    subgroups: dict[Key, DataclassT | type[DataclassT] | functools.partial[DataclassT]],
    *args,
    default: _MISSING_TYPE = MISSING,
    default_factory: _MISSING_TYPE = MISSING,
    **kwargs,
) -> DataclassT:
    ...


def subgroups(
    subgroups: dict[Key, DataclassT | type[DataclassT] | functools.partial[DataclassT]],
    *args,
    default: Key | DataclassT | _MISSING_TYPE = MISSING,
    default_factory: type[DataclassT] | functools.partial[DataclassT] | _MISSING_TYPE = MISSING,
    **kwargs,
) -> DataclassT:
    """Creates a field that will be a choice between different subgroups of arguments.

    This is different than adding a subparser action. There can only be one subparser action, while
    there can be arbitrarily many subgroups. Subgroups can also be nested!


    Parameters
    ----------
    subgroups :
        Dictionary mapping from the subgroup name to the subgroup type.
    default :
        The default subgroup to use, by default MISSING, in which case a subgroup has to be
        selected. Needs to be a key in the subgroups dictionary.
    default_factory :
        The default_factory to use to create the subgroup. Needs to be a value of the `subgroups`
        dictionary.

    Returns
    -------
    A field whose type is the Union of the different possible subgroups.
    """
    if default_factory is not MISSING and default is not MISSING:
        raise ValueError("Can't pass both default and default_factory!")
    from collections.abc import Hashable

    if is_dataclass_instance(default):
        if not isinstance(default, Hashable):
            raise ValueError(
                "'default' can either be a key of the subgroups dict or a hashable (frozen) "
                "dataclass."
            )
        if default not in subgroups.values():
            # TODO: (@lebrice): Do we really need to enforce this? What is the reasoning behind this
            # restriction again?
            raise ValueError(f"Default value {default} needs to be a value in the subgroups dict.")
    elif default is not MISSING and default not in subgroups.keys():
        raise ValueError("default must be a key in the subgroups dict!")

    if default_factory is not MISSING and default_factory not in list(subgroups.values()):
        # NOTE: This is because we need to have a "default key" to associate with the
        # default_factory (and we set that as the default value for the argument of this field).
        raise ValueError("`default_factory` must be a value in the subgroups dict.")
    # IDEA: We could add a `default` key for this `default_factory` value into the `subgroups`
    # dict? However if it's a lambda expression, then we wouldn't then be able to inspect the
    # return type of that default factory (see above). Therefore there doesn't seem to be any
    # good way to allow lambda expressions as default factories yet. Perhaps I'm
    # overcomplicating things and it's actually very simple to do. I'll have to think about it.

    metadata = kwargs.pop("metadata", {})
    metadata["subgroups"] = subgroups
    metadata["subgroup_default"] = default
    metadata["subgroup_dataclass_types"] = {}

    subgroup_dataclass_types: dict[Key, type[DataclassT]] = {}
    choices = subgroups.keys()

    # NOTE: Perhaps we could raise a warning if the default_factory is a Lambda, since we have to
    # instantiate that value in order to inspect the attributes and its values..

    # NOTE: This needs to be the right frame where the subgroups are set.
    _current_frame = inspect.currentframe()
    caller_frame = _current_frame.f_back if _current_frame else None
    for subgroup_key, subgroup_value in subgroups.items():
        if is_lambda(subgroup_value):
            raise NotImplementedError(
                f"Lambda expressions like {subgroup_value!r} can't currently be used as subgroup "
                "values, since we're unable to inspect which dataclass they return without "
                "invoking them.\n"
                "If you want to choose between different versions of a dataclass where arguments "
                "change between subgroups, consider using a `functools.partial` instead. "
            )

        if is_dataclass_instance(subgroup_value):
            dataclass_type = type(subgroup_value)
        elif is_dataclass_type(subgroup_value):
            # all good! Just use that dataclass.
            dataclass_type = subgroup_value
        else:
            try:
                dataclass_type = _get_dataclass_type_from_callable(
                    subgroup_value, caller_frame=caller_frame
                )
            except Exception as exc:
                raise NotImplementedError(
                    f"We are unable to figure out the dataclass to use for the selected subgroup "
                    f"{subgroup_key!r}, because the subgroup value is "
                    f"{subgroup_value!r}, and we don't know what type of "
                    f"dataclass it produces without invoking it!\n"
                    "🙏 Please make an issue on GitHub! 🙏\n"
                    f"Exception raised:\n" + str(exc)
                ) from exc

        subgroup_dataclass_types[subgroup_key] = dataclass_type
    metadata["subgroup_dataclass_types"] = subgroup_dataclass_types

    # TODO: Show the default value from the default factory in the help text.
    # default_factory_dataclass = None
    # if default_factory is not MISSING:
    #     default_factory_dataclass = _get_dataclass_type_from_callable(default_factory)
    # subgroup_field_values = {}

    if default is not MISSING:
        if is_dataclass_instance(default):
            assert default in subgroups.values()
            subgroup_key = [k for k, v in subgroups.items() if v is default][0]
            metadata["subgroup_default"] = subgroup_key
            default = subgroup_key
        else:
            assert default in subgroups.keys()
            default_factory = subgroups[default]
            metadata["subgroup_default"] = default
            default = MISSING

    elif default_factory is not MISSING:
        # assert default_factory in subgroups.values()
        # default_factory passed, which is in the subgroups dict. Find the matching key.
        matching_keys = [k for k, v in subgroups.items() if v is default_factory]
        if not matching_keys:
            # Use == instead of `is` this time.
            matching_keys = [k for k, v in subgroups.items() if v == default_factory]

        # We wouldn't get here if default_factory wasn't in the subgroups dict values.
        assert matching_keys
        if len(matching_keys) > 1:
            raise ValueError(
                f"Default subgroup {default} is found more than once in the subgroups dict?"
            )
        subgroup_default = matching_keys[0]
        metadata["subgroup_default"] = subgroup_default
    else:
        # Store `MISSING` as the subgroup default.
        metadata["subgroup_default"] = MISSING

    from .fields import choice

    return choice(choices, *args, default=default, default_factory=default_factory, metadata=metadata, **kwargs)  # type: ignore


def _get_dataclass_type_from_callable(
    dataclass_fn: Callable[..., DataclassT], caller_frame: inspect.FrameType | None = None
) -> type[DataclassT]:
    """Inspects and returns the type of dataclass that the given callable will return."""
    if is_dataclass_type(dataclass_fn):
        return dataclass_fn

    signature = inspect.signature(dataclass_fn)

    if isinstance(dataclass_fn, functools.partial):
        if is_dataclass_type(dataclass_fn.func):
            return dataclass_fn.func
        # partial to a function that should return a dataclass. Hopefully it has a return type
        # annotation, otherwise we'd have to call the function just to know the return type!
        # NOTE: recurse here, so it also works with `partial(partial(...))` and `partial(some_function)`
        return _get_dataclass_type_from_callable(
            dataclass_fn=dataclass_fn.func, caller_frame=caller_frame
        )

    if signature.return_annotation is inspect.Signature.empty:
        raise TypeError(
            f"Unable to determine what type of dataclass would be returned by the callable "
            f"{dataclass_fn!r}, because it doesn't have a return type annotation, and we don't "
            f"want to call it just to figure out what it produces."
        )
        # NOTE: recurse here, so it also works with `partial(partial(...))` and `partial(some_function)`
        # Recurse, so this also works with partial(partial(...)) (idk why you'd do that though.)

    if isinstance(signature.return_annotation, str):

        dataclass_fn_type = signature.return_annotation
        if caller_frame is not None:
            # Travel up until we find the right frame where the subgroup is defined.

            while (
                caller_frame.f_back is not None
                # TODO: This will stop if it finds a variable with that name! But what if that
                # isn't actually a dataclass ?
                and signature.return_annotation not in caller_frame.f_locals
                and signature.return_annotation not in caller_frame.f_globals
            ):
                caller_frame = caller_frame.f_back

            caller_locals = caller_frame.f_locals
            caller_globals = caller_frame.f_globals

            try:
                # NOTE: This doesn't seem to be very often different than just calling `get_type_hints`
                type_hints = typing.get_type_hints(
                    dataclass_fn, globalns=caller_globals, localns=caller_locals
                )
            except NameError:
                assert False, (caller_locals, caller_globals, caller_frame)
            # assert type_hints == typing.get_type_hints(dataclass_fn)
        else:
            type_hints = typing.get_type_hints(dataclass_fn)
        dataclass_fn_type = type_hints["return"]

        # Recursing here would be a bit extra, let's be real. Might be good enough to just assume that
        # the return annotation needs to be a dataclass.
        # return _get_dataclass_type_from_callable(dataclass_fn_type, caller_frame=caller_frame)
        assert is_dataclass_type(dataclass_fn_type)
        return dataclass_fn_type


def is_lambda(obj: Any) -> bool:
    """Returns True if the given object is a lambda expression.

    Taken from https://stackoverflow.com/questions/3655842/how-can-i-test-whether-a-variable-holds-a-lambda
    """
    LAMBDA = lambda: 0  # noqa: E731
    return isinstance(obj, type(LAMBDA)) and obj.__name__ == LAMBDA.__name__



from __future__ import annotations

import collections
import dataclasses
import functools
import inspect
import typing
from typing import Any, Callable, NamedTuple

import docstring_parser as dp

from . import helpers, parsing


class _Field(NamedTuple):
    name: str
    annotation: type
    field: dataclasses.Field


def _description_from_docstring(docstring: dp.Docstring) -> str:
    """Construct a description from the short and long description of a docstring."""
    description = ""
    if docstring.short_description:
        description += f"{docstring.short_description}\n"
        if docstring.blank_after_short_description:
            description += "\n"
    if docstring.long_description:
        description += f"{docstring.long_description}\n"
        if docstring.blank_after_long_description:
            description += "\n"
    return description


@typing.overload
def main(original_function: None = None, **sp_kwargs) -> Callable[..., Callable[..., Any]]:
    ...


@typing.overload
def main(original_function: Callable[..., Any], **sp_kwargs) -> Callable[..., Any]:
    ...


def main(original_function=None, **sp_kwargs):
    """Parse a function's arguments using simple-parsing from type annotations."""

    def _decorate_with_cli_args(function: Callable[..., Any]) -> Callable[..., Any]:
        """Decorate `function` by binding its arguments obtained from simple-parsing."""

        @functools.wraps(function)
        def _wrapper(*other_args, **other_kwargs) -> Any:
            # Parse signature and parameters
            signature = inspect.signature(function, follow_wrapped=True)
            parameters = signature.parameters

            # Parse docstring to use as help strings
            docstring = dp.parse(inspect.getdoc(function) or "")
            docstring_param_description = {
                param.arg_name: param.description for param in docstring.params
            }

            # Parse all arguments from the function
            fields = []
            for name, parameter in parameters.items():
                # Replace empty annotation with Any
                if parameter.annotation == inspect.Parameter.empty:
                    parameter = parameter.replace(annotation=Any)

                # Parse default or default_factory if the default is callable.
                default, default_factory = dataclasses.MISSING, dataclasses.MISSING
                if parameter.default != inspect.Parameter.empty:
                    if inspect.isfunction(parameter.default):
                        default_factory = parameter.default
                    else:
                        default = parameter.default

                field = _Field(
                    name,
                    parameter.annotation,
                    helpers.field(
                        name=name,
                        default=default,
                        default_factory=default_factory,
                        help=docstring_param_description.get(name, ""),
                        positional=parameter.kind == inspect.Parameter.POSITIONAL_ONLY,
                    ),
                )
                fields.append(field)

            # We can have positional arguments with no defaults that come out of order
            # when parsing the function signature. Therefore, before we construct
            # the dataclass we have to sort fields according to their default value.
            # We query fields by name so there's no need to worry about the order.
            def _field_has_default(field: _Field) -> bool:
                return (
                    field.field.default is not dataclasses.MISSING
                    or field.field.default_factory is not dataclasses.MISSING
                )

            fields = sorted(fields, key=_field_has_default)

            # Create the dataclass using the fields derived from the function's signature
            FunctionArgs = dataclasses.make_dataclass(function.__qualname__, fields)
            FunctionArgs.__doc__ = _description_from_docstring(docstring) or None
            function_args = parsing.parse(
                FunctionArgs,
                dest="args",
                add_config_path_arg=False,
                **sp_kwargs,
            )

            # Construct both positional and keyword arguments.
            args, kwargs = [], {}
            for field in dataclasses.fields(function_args):
                value = getattr(function_args, field.name)
                if field.metadata.get("positional", False):
                    args.append(value)
                else:
                    # TODO: py39: use union operator (|=)
                    kwargs.update({field.name: value})

            # Construct positional arguments with CLI and runtime args
            positionals = (*args, *other_args)
            # Construct keyword arguments so it can override arguments
            # so we don't receive multiple value errors.
            keywords = collections.ChainMap(kwargs, other_kwargs)

            # Call the function
            return function(*positionals, **keywords)

        return _wrapper

    if original_function:
        return _decorate_with_cli_args(original_function)

    return _decorate_with_cli_args



from __future__ import annotations

import dataclasses
from typing import Any, overload

from simple_parsing.utils import DataclassT, is_dataclass_instance, unflatten_split


@overload
def replace(obj: DataclassT, changes_dict: dict[str, Any]) -> DataclassT:
    ...


@overload
def replace(obj: DataclassT, **changes) -> DataclassT:
    ...


def replace(obj: DataclassT, changes_dict: dict[str, Any] | None = None, **changes) -> DataclassT:
    """Replace some values in a dataclass.

    Compared to `dataclasses.replace`, this has two major differences:
    1. Allows recursively replacing values in nested dataclass fields. When a dictionary is passed
    as the value, and the value of that field on `obj` is a dataclass, then `replace` is called
    recursively.

    2. Allows passing a dictionary of flattened changes, e.g. `{"a.b": 1}` instead of
    `{"a": {"b": 1}}`.

    ## Examples

    >>> import dataclasses
    >>> import simple_parsing
    >>> from typing import Union
    >>> @dataclasses.dataclass
    ... class A:
    ...    a: int = 0
    >>> @dataclasses.dataclass
    ... class B:
    ...     b: str = "b"
    >>> @dataclasses.dataclass
    ... class Config:
    ...     a_or_b: Union[A, B] = dataclasses.field(default_factory=A)

    >>> base_config = Config(a_or_b=A(a=1))


    NOTE: key difference with respect to `dataclasses.replace`:
    >>> dataclasses.replace(base_config, a_or_b={"a": 123})
    Config(a_or_b={'a': 123})
    >>> simple_parsing.replace(base_config, a_or_b={"a": 123})
    Config(a_or_b=A(a=123))

    Replace accepts either a dictionary of changes or keyword arguments:

    >>> simple_parsing.replace(base_config, {"a_or_b": B(b="bob")})
    Config(a_or_b=B(b='bob'))
    >>> simple_parsing.replace(base_config, a_or_b=B(b='bob'))
    Config(a_or_b=B(b='bob'))

    Changes can also be passed in a 'flat' format, which makes it easy to replace nested fields:
    >>> simple_parsing.replace(base_config, {"a_or_b.a": 2})
    Config(a_or_b=A(a=2))
    """
    if changes_dict and changes:
        raise ValueError("Cannot pass both `changes_dict` and `changes`")
    changes = changes_dict or changes
    # changes can be given in a 'flat' format in `changes_dict`, e.g. {"a.b.c": 123}.
    # Unflatten them back to a nested dict (e.g. {"a": {"b": {"c": 123}}})
    changes = unflatten_split(changes)

    replace_kwargs = {}
    for field in dataclasses.fields(obj):
        if field.name not in changes:
            continue
        if not field.init:
            raise ValueError(f"Cannot replace value of non-init field {field.name}.")

        field_value = getattr(obj, field.name)

        if is_dataclass_instance(field_value) and isinstance(changes[field.name], dict):
            field_changes = changes.pop(field.name)
            new_value = replace(field_value, **field_changes)
        else:
            new_value = changes.pop(field.name)
        replace_kwargs[field.name] = new_value

    # note: there may be some leftover values in `changes` that are not fields of this dataclass.
    # we still pass those.
    replace_kwargs.update(changes)

    return dataclasses.replace(obj, **replace_kwargs)



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

from .testutils import TestSetup


def test_optional_union():
    @dataclass
    class Config(TestSetup):
        path: Optional[Union[Path, str]] = None

    config = Config.setup("")
    assert config.path is None

    config = Config.setup("--path")
    assert config.path is None

    config = Config.setup("--path bob")
    assert config.path == Path("bob")



"""Simple, Elegant Argument parsing.
@author: Fabrice Normandin
"""
from . import helpers, utils, wrappers
from .conflicts import ConflictResolution
from .decorators import main
from .help_formatter import SimpleHelpFormatter
from .helpers import (
    Partial,
    Serializable,
    choice,
    config_for,
    field,
    flag,
    list_field,
    mutable_field,
    subgroups,
    subparsers,
)
from .parsing import (
    ArgumentGenerationMode,
    ArgumentParser,
    DashVariant,
    NestedMode,
    ParsingError,
    parse,
    parse_known_args,
)
from .replace import replace
from .utils import InconsistentArgumentError

__all__ = [
    "ArgumentGenerationMode",
    "ArgumentParser",
    "choice",
    "config_for",
    "ConflictResolution",
    "DashVariant",
    "field",
    "flag",
    "helpers",
    "InconsistentArgumentError",
    "list_field",
    "main",
    "mutable_field",
    "NestedMode",
    "parse_known_args",
    "parse",
    "ParsingError",
    "Partial",
    "replace",
    "Serializable",
    "SimpleHelpFormatter",
    "subgroups",
    "subparsers",
    "utils",
    "wrappers",
]

from . import _version

__version__ = _version.get_versions()["version"]
